from keras_cv_attention_models.beit.beit import (
    Beit,
    BeitBasePatch16,
    BeitLargePatch16,
    BeitV2BasePatch16,
    BeitV2LargePatch16,
    PatchConv2DWithResampleWeights,
    MultiHeadRelativePositionalEmbedding,
    keras_model_load_weights_from_pytorch_model,
)
from keras_cv_attention_models.beit.flexivit import ViT, FlexiViTSmall, FlexiViTBase, FlexiViTLarge
from keras_cv_attention_models.beit.eva import EvaLargePatch14, EvaGiantPatch14

__beit_head_doc__ = """
Keras implementation of [beit](https://github.com/microsoft/unilm/tree/master/beit).
Paper [PDF 2106.08254 BEIT: BERT Pre-Training of Image Transformers](https://arxiv.org/pdf/2106.08254.pdf).
"""

__beitv2_head_doc__ = """
Keras implementation of [Github microsoft/beit2](https://github.com/microsoft/unilm/tree/master/beit2).
Paper [PDF 2208.06366 BEiT v2: Masked Image Modeling with Vector-Quantized Visual Tokenizers](https://arxiv.org/pdf/2208.06366.pdf).
"""

__vit_head_doc__ = """
Keras implementation of [Github google-research/vision_transformer](https://github.com/google-research/vision_transformer).
Paper [PDF 2010.11929 An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale](https://arxiv.org/pdf/2010.11929.pdf).
"""

__flexivit_head_doc__ = """
Keras implementation of [Github google-research/big_vision](https://github.com/google-research/big_vision/tree/main/big_vision/configs/proj/flexivit).
Paper [PDF 2212.08013 FlexiViT: One Model for All Patch Sizes](https://arxiv.org/pdf/2212.08013.pdf).
"""

__eva_head_doc__ = """
Keras implementation of [Github baaivision/EVA](https://github.com/baaivision/EVA).
Paper [PDF 2211.07636 EVA: Exploring the Limits of Masked Visual Representation Learning at Scale](https://arxiv.org/pdf/2211.07636.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet21k-ft1k' (pre-training on ImageNet21k and fine-tuned ImageNet).
      Will try to download and load pre-trained model weights if not None.
  force_reload_mismatch: boolean value, set True if `patch_size` changed, will force reloading `pos_emb` and `stem_conv` weights.

Returns:
    A `keras.Model` instance.
"""

__class_tail_doc__ = """
Args:
  depth: number of blocks. Default `12`.
  embed_dim: channel dimension for stem and all blocks. Default `768`.
  num_heads: heads number for transformer blocks. Default `12`.
  mlp_ratio: dimension expansion ration for `mlp_block`s. Default `4`.
  patch_size: stem patch size. Default `16`.
  attn_key_dim: key dimension for transformer blocks. Default `0`.
  attn_qv_bias: boolean value, if True and attn_qkv_bias being False, will add BiasLayer for `query` and `value` in transformer.
      Default False for Vit, True for Beit.
  attn_qkv_bias: boolean value, if True will just use bias in `qkv_dense`, and set `qv_bias` False.
      Default True for Vit, False for Beit.
  attn_out_weight: boolean value if use output dense for transformer. Default `True`.
  attn_out_bias: boolean value if output dense use bias for transformer. Default `True`.
  attn_dropout: `attention_score` dropout rate. Default `0`.
  gamma_init_value: init value for `attention` and `mlp` branch `gamma`, if > 0 will use `layer_scale` on block output.
      Default 0 for Vit, 0.1 for Beit
  use_abs_pos_emb: boolean value if use abcolute positional embedding or relative one in attention blocks.
      Default True for Vit, False for Beit.
  use_abs_pos_emb_on_cls_token: boolean value, if `use_abs_pos_emb` is True, whether apply `pos_emb` on `cls_token`.
      False for `FlexiViT`, same as `no_embed_class` in timm. Default True for others.
  use_mean_pooling: boolean value if use mean output or `class_token` output. Default False for Vit, True for Beit.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model                 | Params  | FLOPs   | Input | Top1 Acc |
  | --------------------- | ------- | ------- | ----- | -------- |
  | BeitBasePatch16, 21k  | 86.53M  | 17.61G  | 224   | 85.240   |
  |                       | 86.74M  | 55.70G  | 384   | 86.808   |
  | BeitLargePatch16, 21k | 304.43M | 61.68G  | 224   | 87.476   |
  |                       | 305.00M | 191.65G | 384   | 88.382   |
  |                       | 305.67M | 363.46G | 512   | 88.584   |

  | Model              | Params  | FLOPs  | Input | Top1 Acc |
  | ------------------ | ------- | ------ | ----- | -------- |
  | BeitV2BasePatch16  | 86.53M  | 17.61G | 224   | 85.5     |
  | - imagenet21k-ft1k | 86.53M  | 17.61G | 224   | 86.5     |
  | BeitV2BasePatch16  | 304.43M | 61.68G | 224   | 87.3     |
  | - imagenet21k-ft1k | 304.43M | 61.68G | 224   | 88.4     |

  | Model         | Params  | FLOPs  | Input | Top1 Acc |
  | ------------- | ------- | ------ | ----- | -------- |
  | FlexiViTSmall | 22.06M  | 5.36G  | 240   | 82.53    |
  | FlexiViTBase  | 86.59M  | 20.33G | 240   | 84.66    |
  | FlexiViTLarge | 304.47M | 71.09G | 240   | 85.64    |

  | Model                 | Params  | FLOPs    | Input | Top1 Acc |
  | --------------------- | ------- | -------- | ----- | -------- |
  | EvaLargePatch14, 22k  | 304.14M | 61.65G   | 196   | 88.59    |
  |                       | 304.53M | 191.55G  | 336   | 89.20    |
  | EvaGiantPatch14, clip | 1012.6M | 267.40G  | 224   | 89.10    |
  | - m30m                | 1013.0M | 621.45G  | 336   | 89.57    |
  | - m30m                | 1014.4M | 1911.61G | 560   | 89.80    |
"""

Beit.__doc__ = __beit_head_doc__ + __class_tail_doc__
ViT.__doc__ = __vit_head_doc__ + __class_tail_doc__

__model_tail_doc__ = """
Args:
""" + __tail_doc__

BeitBasePatch16.__doc__ = __beit_head_doc__ + __model_tail_doc__
BeitLargePatch16.__doc__ = BeitBasePatch16.__doc__

BeitV2BasePatch16.__doc__ = __beitv2_head_doc__ + __model_tail_doc__
BeitV2LargePatch16.__doc__ = BeitV2BasePatch16.__doc__

FlexiViTSmall.__doc__ = __flexivit_head_doc__ + __model_tail_doc__
FlexiViTBase.__doc__ = FlexiViTSmall.__doc__
FlexiViTLarge.__doc__ = FlexiViTSmall.__doc__

EvaLargePatch14.__doc__ = __eva_head_doc__ + __model_tail_doc__
EvaGiantPatch14.__doc__ = EvaLargePatch14.__doc__

MultiHeadRelativePositionalEmbedding.__doc__ = __beit_head_doc__ + """
Multi Head Relative Positional Embedding layer.

Positional embedding shape is `[num_heads, (2 * height - 1) * (2 * width - 1)]`.
input (with_cls_token=True): `[batch, num_heads, attn_blocks, attn_blocks]`. where `attn_blocks = attn_height * attn_width + class_token`
input (with_cls_token=False): `[batch, num_heads, attn_blocks, attn_blocks]`. where `attn_blocks = attn_height * attn_width`
output: `[batch, num_heads, attn_blocks, attn_blocks] + positional_bias`.
conditions: attn_height == attn_width

Args:
  with_cls_token: boolean value if input is with class_token.
  attn_height: specify `height` for `attn_blocks` if not square `attn_height != attn_width`.
  num_heads: specify num_heads, or using `input.shape[1]`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.MultiHeadRelativePositionalEmbedding()
>>> print(f"{aa(tf.ones([1, 8, 29 * 29 + 1, 29 * 29 + 1])).shape = }")
# aa(tf.ones([1, 8, 29 * 29 + 1, 29 * 29 + 1])).shape = TensorShape([1, 8, 842, 842])
>>> print({ii.name:ii.numpy().shape for ii in aa.weights})
# {'multi_head_relative_positional_embedding/pos_emb:0': (3252, 8)}

>>> aa = attention_layers.MultiHeadRelativePositionalEmbedding(with_cls_token=False)
>>> print(f"{aa(tf.ones([1, 8, 29 * 29, 29 * 29])).shape = }")
# aa(tf.ones([1, 8, 29 * 29, 29 * 29])).shape = TensorShape([1, 8, 841, 841])
>>> print({ii.name:ii.numpy().shape for ii in aa.weights})
# {'multi_head_relative_positional_embedding_1/pos_emb:0': (3249, 8)}

>>> plt.imshow(aa.relative_position_index)
"""

PatchConv2DWithResampleWeights.__doc__ = __flexivit_head_doc__ + """
Source implementation https://github.com/google-research/big_vision/blob/main/big_vision/models/proj/flexi/vit.py#L30
Typical Conv2D layer with `load_resized_weights` function, also using PyTorch padding style.

Args: Same with Conv2D.
"""

keras_model_load_weights_from_pytorch_model.__doc__ = """
Keras ViT model reload weights from timm torch model using `keras_cv_attention_models.download_and_load.keras_reload_from_torch_model`.
Can be used for loading most `Beit` / `ViT` / `FlexViT` / `EVA` models.
Parameters in this using of `keras_reload_from_torch_model` is specifically fitted for timm ViT models.

Args:
  keras_model: keras model same architecture with `timm_vit_model`.
      Need to set parameters for `keras_cv_attention_models.vit.ViT` same as torch model.
  timm_vit_model: built timm vit model.
  save_name: output file name. If ends with `.h5` will save a h5 model, or will be saved_model format.
      Default None for using `{model_name}_{input_shape}.h5`.

Outputs:
  Converted `save_name` weight file, which can be used for afterward and further usage.

Examples:
>>> # Build a timm ViT model
>>> import timm
>>> torch_model = timm.models.vit_tiny_patch16_224(pretrained=True)
>>> _ = torch_model.eval()
>>>
>>> # Build a ViT model same architecture with torch_model
>>> from keras_cv_attention_models import vit
>>> mm = vit.ViT(depth=12, embed_dim=192, num_heads=3, pretrained=None, classifier_activation=None)
>>> vit.keras_model_load_weights_from_pytorch_model(mm, torch_model)
# >>>> Save model to: vit_224.h5
# >>>> Keras model prediction: [('n02123045', 'tabby', 11.990417), ('n02123159', 'tiger_cat', 11.630723), ...]
# >>>> Torch model prediction: [[('n02123045', 'tabby', 11.99042), ('n02123159', 'tiger_cat', 11.630725), ...]
"""



from keras_cv_attention_models.pvt.pvt import PyramidVisionTransformerV2, PVT_V2B0, PVT_V2B1, PVT_V2B2, PVT_V2B2_linear, PVT_V2B3, PVT_V2B4, PVT_V2B5

__head_doc__ = """
Keras implementation of [Github whai362/PVT](https://github.com/whai362/PVT/tree/v2/classification).
Paper [PDF 2106.13797 PVTv2: Improved Baselines with Pyramid Vision Transformer](https://arxiv.org/pdf/2106.13797.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  layer_scale: layer scale init value, [Going deeper with Image Transformers](https://arxiv.org/abs/2103.17239).
      Default 0 for not using.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet21k-ft1k' (pre-training on ImageNet21k and fine-tuned ImageNet).
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

PyramidVisionTransformerV2.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  embed_dims: output channels for each stack.
  num_heads: int or list value indicates heads number for transformer blocks in each stack.
  mlp_ratios: int or list value indicates expand ratio for mlp blocks hidden channel in each stack.
  sr_ratios: int or list value indicates attention blocks key_value downsample rate in each stack.
  stem_patch_size: stem patch size. Default `7`.
  use_linear: boolean value if using linear complexity attention layer with `AvgPool2D`. True for `PVT_V2B2_linear`.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model           | Params | FLOPs  | Input | Top1 Acc |
  | --------------- | ------ | ------ | ----- | -------- |
  | PVT_V2B0        | 3.7M   | 580.3M | 224   | 70.5     |
  | PVT_V2B1        | 14.0M  | 2.14G  | 224   | 78.7     |
  | PVT_V2B2        | 25.4M  | 4.07G  | 224   | 82.0     |
  | PVT_V2B2_linear | 22.6M  | 3.94G  | 224   | 82.1     |
  | PVT_V2B3        | 45.2M  | 6.96G  | 224   | 83.1     |
  | PVT_V2B4        | 62.6M  | 10.19G | 224   | 83.6     |
  | PVT_V2B5        | 82.0M  | 11.81G | 224   | 83.8     |
"""

PVT_V2B0.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

PVT_V2B1.__doc__ = PVT_V2B0.__doc__
PVT_V2B2.__doc__ = PVT_V2B0.__doc__
PVT_V2B2_linear.__doc__ = PVT_V2B0.__doc__
PVT_V2B3.__doc__ = PVT_V2B0.__doc__
PVT_V2B4.__doc__ = PVT_V2B0.__doc__
PVT_V2B5.__doc__ = PVT_V2B0.__doc__



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    ChannelAffine,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    layer_norm,
    HeadInitializer,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-6
PRETRAINED_DICT = {
    "convnext_tiny": {
        "imagenet": "1deac703865e190528899d5c489afa37",
        "imagenet21k-ft1k": {224: "b70650cc030ec528802762f58940095d", 384: "d6653ede30e25e0c6240f546675393ad"},
    },
    "convnext_small": {
        "imagenet": "7e75873348d445eb2aab4200a5d49f80",
        "imagenet21k-ft1k": {224: "da7c257650b112c1537f2753166fae49", 384: "37ff23f51f2ec9d9b6de2ea7d732ac5f"},
    },
    "convnext_base": {
        "imagenet": {224: "dddac5dcd13bffc1e05688f529726f8c", 384: "ae8dc9bbca6472dc12de30db95ea1018"},
        "imagenet21k-ft1k": {224: "40f78cec6cd327392a9d24f968f9e76b", 384: "4829ff932a930117525920317083d317"},
    },
    "convnext_large": {
        "imagenet": {224: "32d401c254b623d36c22f232884000ba", 384: "01b4e72ca589c2f0ac15551e06d29818"},
        "imagenet21k-ft1k": {224: "dc211e955875f8ab6de7518253e41a46", 384: "68ef87754d6ca634e32d2326c34ddd0b"},
    },
    "convnext_xlarge": {"imagenet21k-ft1k": {224: "7c7ab46f41ac34655f3e035b873a2163", 384: "636db850c0a73ba10e8ab32e91c38df6"}},
    "convnext_v2_atto": {"imagenet": "e604fa1edfefe6207957feec4f5612db"},
    "convnext_v2_base": {
        "imagenet": "879caa3189ed74ed969f9348b82afe47",
        "imagenet21k-ft1k": {224: "8d15a1e29f28e3fd8f0e6691e872ebee", 384: "b267df29706944ec4bc60b57c9778be0"},
    },
    "convnext_v2_femto": {"imagenet": "46d4e39a2efb4dc0aa543442b9000d89"},
    "convnext_v2_huge": {
        "imagenet": "347d28c6354964c30a04c5f6cadf0ebc",
        "imagenet21k-ft1k": {384: "dfad27a621300ae254ff812827a03354", 512: "0b40599908e70e42e32c2a206f94abf3"},
    },
    "convnext_v2_large": {
        "imagenet": "18327817424ada5a1c4ea257079e0694",
        "imagenet21k-ft1k": {224: "4bce3ade2680d7181c782b65df8ed929", 384: "6d01f83513538e1f02640314e044d00e"},
    },
    "convnext_v2_nano": {
        "imagenet": "32911de07188225277a47219dbdb4134",
        "imagenet21k-ft1k": {224: "e1761b343263167eb9f4d6c33c6c892d", 384: "2980c5a37ad16cfbc6c90b8a8bb1c83f"},
    },
    "convnext_v2_pico": {"imagenet": "27ed3ae499e0ca6f6b5e3cf8e041ab92"},
    "convnext_v2_tiny": {
        "imagenet": "4b0a70c87400993385b668853c4e3654",
        "imagenet21k-ft1k": {224: "de1db9ab2d8c565767cf81401ceed6ae", 384: "cc9028f2baa22ac1799ca1219e7b2991"},
    },
}


def global_response_normalize(inputs, axis="auto", name=None):
    axis = (-1 if backend.image_data_format() == "channels_last" else 1) if axis == "auto" else axis
    num_dims = len(inputs.shape)
    axis = num_dims + axis if axis < 0 else axis
    nn = functional.norm(inputs, axis=[ii for ii in range(1, num_dims) if ii != axis], keepdims=True)
    nn = nn / (functional.reduce_mean(nn, axis=axis, keepdims=True) + 1e-6)
    nn = ChannelAffine(use_bias=True, weight_init_value=0, axis=axis, name=name and name + "gamma")(inputs * nn)
    return nn + inputs


def add_with_layer_scale_and_drop_block(short, deep, layer_scale=0, residual_scale=0, drop_rate=0, axis="auto", name=""):
    """Just simplify calling, perform `out = short + drop_block(layer_scale(deep))`"""
    axis = (-1 if backend.image_data_format() == "channels_last" else 1) if axis == "auto" else axis
    short = ChannelAffine(use_bias=False, weight_init_value=residual_scale, axis=axis, name=name + "res_gamma")(short) if residual_scale > 0 else short
    deep = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=axis, name=name + "gamma")(deep) if layer_scale > 0 else deep
    deep = drop_block(deep, drop_rate=drop_rate, name=name)
    # print(f">>>> {short.shape = }, {deep.shape = }")
    return layers.Add(name=name + "output")([short, deep])


def block(inputs, output_channel, layer_scale_init_value=1e-6, use_grn=False, drop_rate=0, activation="gelu", name=""):
    nn = depthwise_conv2d_no_bias(inputs, kernel_size=7, padding="SAME", use_bias=True, name=name)
    nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name=name)

    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute((2, 3, 1), name=name + "permute_pre")(nn)
    nn = layers.Dense(4 * output_channel, name=name + "up_dense")(nn)
    nn = activation_by_name(nn, activation, name=name)
    if use_grn:
        nn = global_response_normalize(nn, axis=-1, name=name + "grn_")  # also using axis=-1 for channels_first
    nn = layers.Dense(output_channel, name=name + "down_dense")(nn)
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute((3, 1, 2), name=name + "permute_post")(nn)

    return add_with_layer_scale_and_drop_block(inputs, nn, layer_scale=layer_scale_init_value, drop_rate=drop_rate, name=name)


def ConvNeXt(
    num_blocks=[3, 3, 9, 3],
    out_channels=[96, 192, 384, 768],
    stem_width=-1,
    layer_scale_init_value=1e-6,  # 1e-6 for v1, 0 for v2
    use_grn=False,  # False for v1, True for v2
    head_init_scale=1.0,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0.1,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="convnext",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ Stem """
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = conv2d_no_bias(inputs, stem_width, kernel_size=4, strides=4, padding="VALID", use_bias=True, name="stem_")
    nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name="stem_")

    """ Blocks """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel) in enumerate(zip(num_blocks, out_channels)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name=stack_name + "downsample_")
            nn = conv2d_no_bias(nn, out_channel, kernel_size=2, strides=2, use_bias=True, name=stack_name + "downsample_")
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = block(nn, out_channel, layer_scale_init_value, use_grn, block_drop_rate, activation, name=block_name)
            global_block_id += 1

    """  Output head """
    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name="head_")
        head_init = HeadInitializer(scale=head_init_scale)
        nn = layers.Dense(
            num_classes, dtype="float32", activation=classifier_activation, kernel_initializer=head_init, bias_initializer=head_init, name="predictions"
        )(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="convnext", pretrained=pretrained)
    return model


def ConvNeXtTiny(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 9, 3]
    out_channels = [96, 192, 384, 768]
    return ConvNeXt(**locals(), model_name="convnext_tiny", **kwargs)


def ConvNeXtSmall(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 27, 3]
    out_channels = [96, 192, 384, 768]
    return ConvNeXt(**locals(), model_name="convnext_small", **kwargs)


def ConvNeXtBase(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 27, 3]
    out_channels = [128, 256, 512, 1024]
    return ConvNeXt(**locals(), model_name="convnext_base", **kwargs)


def ConvNeXtLarge(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 27, 3]
    out_channels = [192, 384, 768, 1536]
    return ConvNeXt(**locals(), model_name="convnext_large", **kwargs)


def ConvNeXtXlarge(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    num_blocks = [3, 3, 27, 3]
    out_channels = [256, 512, 1024, 2048]
    return ConvNeXt(**locals(), model_name="convnext_xlarge", **kwargs)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    add_with_layer_scale_and_drop_block,
    batchnorm_with_activation,
    conv2d_no_bias,
    ChannelAffine,
    depthwise_conv2d_no_bias,
    output_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "moganet_xtiny": {"imagenet": "b1a7a1b77777cd8fdc8b5b5333e49215"},
    "moganet_tiny": {"imagenet": {224: "2ecd6f4552fb5bdbd0845c6a49bb67a9", 256: "6be0f1b79d00ba535412c1d1d3c1a71f"}},
    "moganet_small": {"imagenet": "8b090a8058304bdbacdc70896ecb25cd"},
    "moganet_base": {"imagenet": "3aa0d6fccc312bb7674495b7c0c8153e"},
    "moganet_large": {"imagenet": "1714094064c837afa3c6f0304563b496"},
}


def feature_decompose(inputs, use_pool=False, activation="gelu", name=""):
    channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
    if use_pool:
        decomposed = layers.GlobalAveragePooling2D(keepdims=True, name=name + "pool")(inputs)
    else:
        decomposed = conv2d_no_bias(inputs, 1, use_bias=True, name=name)
    decomposed = activation_by_name(decomposed, activation=activation, name=name)
    decomposed = ChannelAffine(use_bias=False, weight_init_value=1e-5, axis=channel_axis, name=name + "affine")(inputs - decomposed)
    return inputs + decomposed


def multi_order_depthwise_conv2d(inputs, dilations=[1, 2, 3], channel_split=[1, 3, 4], name=""):
    channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]
    channel_split = [-1] + [int(input_channel * ii / sum(channel_split)) for ii in channel_split[1:]]
    paddings = [(1 + 4 * dilations[0]) // 2, (1 + 4 * dilations[1]) // 2, (1 + 6 * dilations[2]) // 2]

    # print(f"{inputs.shape = }, {input_channel = }, {channel_split = }")
    first = depthwise_conv2d_no_bias(inputs, kernel_size=5, padding=paddings[0], use_bias=True, dilation_rate=dilations[0], name=name + "first_")
    first, second, third = functional.split(first, channel_split, axis=channel_axis)
    second = depthwise_conv2d_no_bias(second, kernel_size=5, padding=paddings[1], use_bias=True, dilation_rate=dilations[1], name=name + "second_")
    third = depthwise_conv2d_no_bias(third, kernel_size=7, padding=paddings[2], use_bias=True, dilation_rate=dilations[2], name=name + "third_")
    # print(f"{first.shape = }, {second.shape = }, {third.shape = }")

    out = functional.concat([first, second, third], axis=channel_axis)
    out = conv2d_no_bias(out, input_channel, use_bias=True, name=name + "out_")
    return out


def moga_block(inputs, mlp_ratio=4, layer_scale=0, drop_rate=0, attn_activation="swish", activation="gelu", name=""):
    input_channel = inputs.shape[-1 if backend.image_data_format() == "channels_last" else 1]

    """ attention swish """
    pre_norm = batchnorm_with_activation(inputs, activation=None, name=name + "pre_attn_")
    nn = conv2d_no_bias(pre_norm, input_channel, use_bias=True, name=name + "pre_attn_")  # proj_1
    nn = feature_decompose(nn, use_pool=True, activation=None, name=name + "pre_attn_decompose_")
    nn = activation_by_name(nn, activation=attn_activation, name=name + "pre_attn_")

    gate = conv2d_no_bias(nn, input_channel, use_bias=True, name=name + "attn_gate_")
    gate = activation_by_name(gate, activation=attn_activation, name=name + "attn_gate_")
    value = multi_order_depthwise_conv2d(nn, name=name + "attn_value_")  # MultiOrderDWConv
    value = activation_by_name(value, activation=attn_activation, name=name + "attn_value_")
    gate_value = layers.Multiply(name=name + "attn_gate_value")([gate, value])
    gate_value = conv2d_no_bias(gate_value, input_channel, use_bias=True, name=name + "attn_gate_value_")  # proj_2
    nn = layers.Add(name=name + "attn_out")([pre_norm, gate_value])

    attn_out = add_with_layer_scale_and_drop_block(inputs, nn, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "attn_")

    """ MLP proj 1 """
    nn = batchnorm_with_activation(attn_out, activation=None, name=name + "mlp_")
    nn = conv2d_no_bias(nn, input_channel * mlp_ratio, use_bias=True, name=name + "mlp_1_")
    nn = depthwise_conv2d_no_bias(nn, kernel_size=3, padding="SAME", use_bias=True, name=name + "mlp_2_")
    nn = activation_by_name(nn, activation=activation, name=name + "mlp_")
    # drop

    """ MLP proj 2 """
    nn = feature_decompose(nn, activation=activation, name=name + "mlp_decompose_")
    nn = conv2d_no_bias(nn, input_channel, use_bias=True, name=name + "mlp_3_")
    nn = add_with_layer_scale_and_drop_block(attn_out, nn, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "mlp_")
    return nn


def MogaNet(
    num_blocks=[3, 3, 10, 2],
    out_channels=[32, 64, 96, 192],
    mlp_ratios=[8, 8, 4, 4],
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    attn_activation="swish",
    drop_connect_rate=0,
    dropout=0,
    layer_scale=1e-5,
    classifier_activation="softmax",
    pretrained=None,
    model_name="moganet",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ Stem """
    nn = conv2d_no_bias(inputs, out_channels[0] // 2, kernel_size=3, strides=2, padding="SAME", use_bias=True, name="stem_1_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_1_")
    nn = conv2d_no_bias(nn, out_channels[0], kernel_size=3, strides=2, padding="SAME", use_bias=True, name="stem_2_")
    nn = batchnorm_with_activation(nn, activation=None, name="stem_2_")

    """ stacks """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel) in enumerate(zip(num_blocks, out_channels)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = conv2d_no_bias(nn, out_channel, 3, strides=2, padding="same", use_bias=True, name=stack_name + "downsample_")
            nn = batchnorm_with_activation(nn, activation=None, name=stack_name + "downsample_")

        mlp_ratio = mlp_ratios[stack_id] if isinstance(mlp_ratios, (list, tuple)) else mlp_ratios
        for block_id in range(num_block):
            name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = moga_block(nn, mlp_ratio, layer_scale, block_drop_rate, attn_activation=attn_activation, activation=activation, name=name)
            global_block_id += 1
        nn = batchnorm_with_activation(nn, activation=None, name=stack_name + "output_")

    nn = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "moganet", pretrained)
    return model


def MogaNetXtiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MogaNet(**locals(), model_name="moganet_xtiny", **kwargs)


def MogaNetTiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 12, 2]
    out_channels = [32, 64, 128, 256]
    return MogaNet(**locals(), model_name="moganet_tiny", **kwargs)


def MogaNetSmall(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 3, 12, 2]
    out_channels = [64, 128, 320, 512]
    return MogaNet(**locals(), model_name="moganet_small", **kwargs)


def MogaNetBase(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [4, 6, 22, 3]
    out_channels = [64, 160, 320, 512]
    return MogaNet(**locals(), model_name="moganet_base", **kwargs)


def MogaNetLarge(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [4, 6, 44, 4]
    out_channels = [64, 160, 320, 640]
    return MogaNet(**locals(), model_name="moganet_large", **kwargs)



from keras_cv_attention_models.cmt.cmt import (
    CMT,
    CMT_torch,
    CMTTiny,
    CMTTiny_torch,
    CMTXS_torch,
    CMTSmall_torch,
    CMTBase_torch,
    light_mhsa_with_multi_head_relative_position_embedding,
    BiasPositionalEmbedding,
)

__head_doc__ = """
Keras implementation of CMT.
Paper [PDF 2107.06263 CMT: Convolutional Neural Networks Meet Vision Transformers](https://arxiv.org/pdf/2107.06263.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  dropout: dropout rate if top layers is included.
  pretrained: None or "imagenet". Only CMTTiny pretrained available.
  **kwargs: other parameters if available.

Returns:
    A `keras.Model` instance.
"""

CMT.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  stem_width: output dimension for stem block.
  num_heads: heads number for transformer block.
  sr_ratios: attenstion blocks key_value downsample rate.
  ffn_expansion: IRFFN blocks hidden expansion rate.
  qkv_bias: boolean value if attention block query / key /value using bias. Default False for CMT, True for CMT_torch,
  attn_out_bias: boolean value if attention block output using bias. Default False for CMT, True for CMT_torch,
  attn_use_bn: boolean value if attention block using batchNorm or layerNorm. Default False for CMT, True for CMT_torch,
  use_block_pos_emb: boolean value if using shared positional embeddings for same block.
      Default False for CMT using individual `MultiHeadRelativePositionalEmbedding` in each attention block.
      Default True for CMT_torch using shared `BiasPositionalEmbedding` in attention blocks of same stack.
  feature_activation: None for same with `activation`, or specific activation for feature block.
      Default None for CMT, "swish" for CMT_torch,
  feature_act_first: boolean value if using `activation -> batchnorm` or `batchnorm -> activation` for feature block.
      Default True for CMT, False for CMT_torch,
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model                              | Params | FLOPs | Input | Top1 Acc |
  | ---------------------------------- | ------ | ----- | ----- | -------- |
  | CMTTiny, (Self trained 105 epochs) | 9.5M   | 0.65G | 160   | 77.4     |
  | - 305 epochs                       | 9.5M   | 0.65G | 160   | 78.94    |
  | - fine-tuned 224 (69 epochs)       | 9.5M   | 1.32G | 224   | 80.73    |
  | CMTTiny_torch, 1000 epochs         | 9.5M   | 0.65G | 160   | 79.2     |
  | CMTXS_torch                        | 15.2M  | 1.58G | 192   | 81.8     |
  | CMTSmall_torch                     | 25.1M  | 4.09G | 224   | 83.5     |
  | CMTBase_torch                      | 45.7M  | 9.42G | 256   | 84.5     |
"""
CMT_torch.__doc__ = CMT.__doc__

CMTTiny.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

CMTTiny_torch.__doc__ = CMTTiny.__doc__
CMTXS_torch.__doc__ = CMTTiny.__doc__
CMTSmall_torch.__doc__ = CMTTiny.__doc__
CMTBase_torch.__doc__ = CMTTiny.__doc__

light_mhsa_with_multi_head_relative_position_embedding.__doc__ = __head_doc__ + """
Light multi head self attention with relative position embedding. Defined as function, not layer.
Downsample `key_value` with a `sr_ratio` using `DepthwiseConv2D` + `LayerNorm`.
Also adds `MultiHeadRelativePositionalEmbedding` to `attention_scores`.

Args:
  inputs: input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query, key and value. Default `0` for `key_dim = inputs.shape[-1] // num_heads`.
  sr_ratio: `key_value` downsample rate.
  qkv_bias: Boolean value if attention block query / key /value using bias.
  pos_emb: Specific instantiated positional embedding class, or None for using `MultiHeadRelativePositionalEmbedding`.
  use_bn: Boolean value if attention block using batchNorm or layerNorm.
  out_shape: The expected shape of an output tensor. If not specified, projects back to the input dim.
  out_weight: Boolean, whether use an ouput dense.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  dropout: Dropout probability for attention.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.light_mhsa_with_multi_head_relative_position_embedding(inputs, num_heads=4, sr_ratio=3, out_shape=512, name="attn_")
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 512])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'attn_kv_sr_dw_conv/depthwise_kernel:0': TensorShape([3, 3, 256, 1]),
#  'attn_kv_sr_ln/gamma:0': TensorShape([256]),
#  'attn_kv_sr_ln/beta:0': TensorShape([256]),
#  'attn_query/kernel:0': TensorShape([256, 256]),
#  'attn_key_value/kernel:0': TensorShape([256, 512]),
#  'attn_pos_emb/positional_embedding:0': TensorShape([4, 837]),
#  'attn_output/kernel:0': TensorShape([256, 512])}
"""

BiasPositionalEmbedding.__doc__ = __head_doc__ + """
A basic bias only layer, with `load_resized_weights` method.

input: `[batch, num_heads, query_height * query_width, kv_height * kv_width]`.
       where `kv_height = query_height // sr_ratio, kv_width = query_width // sr_ratio`
output: input + positional_embedding

Args:
  axis: The dimension applying bias.
  attn_height: specify `height` for `attn_blocks` if not square `attn_height != attn_width`.

Examples:
>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.BiasPositionalEmbedding()
>>> print(f"{aa(tf.ones([2, 4, 36, 9])).shape = }")
# aa(tf.ones([2, 4, 36, 9])).shape = TensorShape([2, 4, 36, 9])
>>> print(f"{aa.bb.shape = }")
# aa.bb.shape = TensorShape([4, 36, 9])
"""



import math
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, initializers, image_data_format
from keras_cv_attention_models import model_surgery
from keras_cv_attention_models import efficientnet
from keras_cv_attention_models.attention_layers import activation_by_name, add_pre_post_process
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.coco import eval_func, anchors_func

BATCH_NORM_EPSILON = 1e-3
PRETRAINED_DICT = {
    "efficientdet_d0": {"coco": {512: "d5ba1c9ffd3627571bb00091e3fde28e"}},
    "efficientdet_d1": {"coco": {640: "c34d8603ea1eefbd7f10415a5b51d824"}},
    "efficientdet_d2": {"coco": {768: "bfd7d19cd5a770e478b51c09e2182dc7"}},
    "efficientdet_d3": {"coco": {896: "39d7b5ac515e073123686c200573b7be"}},
    "efficientdet_d4": {"coco": {1024: "b80d45757f139ef27bf5cf8de0746222"}},
    "efficientdet_d5": {"coco": {1280: "e7bc8987e725d8779684da580a42efa9"}},
    "efficientdet_d6": {"coco": {1280: "9302dc59b8ade929ea68bd68fc71c9f3"}},
    "efficientdet_d7": {"coco": {1536: "6d615faf95a4891aec12392f17155950"}},
    "efficientdet_d7x": {"coco": {1536: "a1e4b4e8e488eeabee14c7801028984d"}},
    "efficientdet_lite0": {"coco": {320: "f6d9aeb36eb0377ff135b9d367f8b4a6"}},
    "efficientdet_lite1": {"coco": {384: "7a593f798462e558deeec6ef976574fc"}},
    "efficientdet_lite2": {"coco": {448: "b48fa6246a50d3d59785cc119c52fc94"}},
    "efficientdet_lite3": {"coco": {512: "945e66f31622d2806aeafde4dff3b4b7"}},
    "efficientdet_lite3x": {"coco": {640: "bcd55e31b99616439ad4f988e2337b86"}},
    "efficientdet_lite4": {"coco": {640: "9f5eadec38498faffb8d6777cb09e3a7"}},
}


@backend.register_keras_serializable(package="efficientdet")
class ReluWeightedSum(layers.Layer):
    def __init__(self, initializer="ones", epsilon=1e-4, **kwargs):
        super().__init__(**kwargs)
        self.initializer, self.epsilon = initializer, epsilon

    def build(self, input_shape):
        self.total = len(input_shape)
        self.gain = self.add_weight(name="gain", shape=(self.total,), initializer=self.initializer, dtype="float32", trainable=True)
        self.__epsilon__ = float(self.epsilon)
        super().build(input_shape)

    def call(self, inputs):
        gain = functional.relu(self.gain)
        gain = gain / (functional.reduce_sum(gain) + self.__epsilon__)
        return functional.reduce_sum([inputs[id] * gain[id] for id in range(self.total)], axis=0)

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"initializer": self.initializer, "epsilon": self.epsilon})
        return base_config


def align_feature_channel(inputs, output_channel, name=""):
    # print(f">>>> align_feature_channel: {name = }, {inputs.shape = }, {output_channel = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    nn = inputs
    if inputs.shape[channel_axis] != output_channel:
        nn = layers.Conv2D(output_channel, kernel_size=1, name=name + "channel_conv")(nn)
        nn = layers.BatchNormalization(epsilon=BATCH_NORM_EPSILON, name=name + "channel_bn")(nn)
    return nn


def resample_fuse(inputs, output_channel, use_weighted_sum=True, interpolation="nearest", use_sep_conv=True, activation="swish", name=""):
    inputs[0] = align_feature_channel(inputs[0], output_channel, name=name)

    if use_weighted_sum:
        nn = ReluWeightedSum(name=name + "wsm")(inputs)
    else:
        nn = layers.Add(name=name + "sum")(inputs)
    nn = activation_by_name(nn, activation, name=name)
    if use_sep_conv:
        nn = layers.SeparableConv2D(output_channel, kernel_size=3, padding="SAME", use_bias=True, name=name + "sepconv")(nn)
    else:
        nn = layers.Conv2D(output_channel, kernel_size=3, padding="SAME", use_bias=True, name=name + "conv")(nn)
    nn = layers.BatchNormalization(epsilon=BATCH_NORM_EPSILON, name=name + "bn")(nn)
    return nn


def bi_fpn(features, output_channel, use_weighted_sum=True, use_sep_conv=True, interpolation="nearest", activation="swish", name=""):
    # print(f">>>> bi_fpn: {[ii.shape for ii in features] = }")
    # features: [p3, p4, p5, p6, p7]
    up_features = [features[-1]]
    for id, feature in enumerate(features[:-1][::-1]):
        cur_name = name + "p{}_up_".format(len(features) - id + 1)
        # up_feature = layers.UpSampling2D(size=(2, 2), interpolation=interpolation, name=cur_name + "up")(up_features[-1])
        size = functional.shape(feature)[1:-1] if image_data_format() == "channels_last" else functional.shape(feature)[2:]
        up_feature = functional.resize(up_features[-1], size, method=interpolation)
        up_feature = resample_fuse([feature, up_feature], output_channel, use_weighted_sum, use_sep_conv=use_sep_conv, activation=activation, name=cur_name)
        up_features.append(up_feature)
    # print(f">>>> bi_fpn: {[ii.shape for ii in up_features] = }")

    # up_features: [p7, p6_up, p5_up, p4_up, p3_up]
    out_features = [up_features[-1]]  # [p3_up]
    up_features = up_features[1:-1][::-1]  # [p4_up, p5_up, p6_up]
    for id, feature in enumerate(features[1:]):
        cur_name = name + "p{}_out_".format(len(features) - 1 + id)
        down_feature = layers.MaxPool2D(pool_size=3, strides=2, padding="SAME", name=cur_name + "max_down")(out_features[-1])
        fusion_feature = [feature, down_feature] if id == len(up_features) else [feature, up_features[id], down_feature]
        out_feature = resample_fuse(fusion_feature, output_channel, use_weighted_sum, use_sep_conv=use_sep_conv, activation=activation, name=cur_name)
        out_features.append(out_feature)
    # out_features: [p3_up, p4_out, p5_out, p6_out, p7_out]
    return out_features


def det_header_pre(features, filters, depth, use_sep_conv=True, activation="swish", name=""):
    # print(f">>>> det_header_pre: {[ii.shape for ii in features] = }")
    if use_sep_conv:
        names = [name + "{}_sepconv".format(id + 1) for id in range(depth)]
        convs = [layers.SeparableConv2D(filters, kernel_size=3, padding="SAME", use_bias=True, name=names[id]) for id in range(depth)]
    else:
        names = [name + "{}_conv".format(id + 1) for id in range(depth)]
        convs = [layers.Conv2D(filters, kernel_size=3, padding="SAME", use_bias=True, name=names[id]) for id in range(depth)]

    outputs = []
    for feature_id, feature in enumerate(features):
        nn = feature
        for id in range(depth):
            nn = convs[id](nn)
            cur_name = name + "{}_{}_bn".format(id + 1, feature_id + 1)
            nn = layers.BatchNormalization(epsilon=BATCH_NORM_EPSILON, name=cur_name)(nn)
            nn = activation_by_name(nn, activation, name=cur_name + "{}_".format(id + 1))
        outputs.append(nn)
    return outputs


def det_header_post(inputs, classes=80, anchors=9, bias_init="zeros", use_sep_conv=True, head_activation="sigmoid", name=""):
    if use_sep_conv:
        header_conv = layers.SeparableConv2D(classes * anchors, kernel_size=3, padding="SAME", bias_initializer=bias_init, name=name + "head")
    else:
        header_conv = layers.Conv2D(classes * anchors, kernel_size=3, padding="SAME", bias_initializer=bias_init, name=name + "conv_head")
    outputs = [header_conv(ii) for ii in inputs]
    outputs = outputs if image_data_format() == "channels_last" else [layers.Permute([2, 3, 1])(ii) for ii in outputs]
    outputs = [layers.Reshape([-1, classes])(ii) for ii in outputs]
    outputs = functional.concat(outputs, axis=1)
    outputs = activation_by_name(outputs, head_activation, name=name + "output_")
    return outputs


def EfficientDet(
    backbone,
    features_pick=[-3, -2, -1],  # The last 3 ones, or specific layer_names / feature_indexes
    additional_features=2,  # Add p5->p6, p6->p7
    fpn_depth=3,
    head_depth=3,
    num_channels=64,
    use_weighted_sum=True,
    anchors_mode="efficientdet",
    use_object_scores="auto",  # "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False
    num_anchors="auto",  # "auto" means: anchors_mode=="anchor_free" -> 1, anchors_mode=="yolor" -> 3, else 9
    num_classes=90,
    use_sep_conv=True,
    activation="swish",
    classifier_activation="sigmoid",
    freeze_backbone=False,
    pretrained="coco",
    model_name=None,
    pyramid_levels_min=3,  # Init anchors for model prediction, not for model structure
    anchor_scale="auto",  # Init anchors for model prediction. "auto" means 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4
    rescale_mode="torch",  # Model precessing input, not for model structure
    input_shape=None,  # Not using, recieving parameter
    kwargs=None,  # Not using, recieving parameter
):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    backbone.trainable = False if freeze_backbone else True
    use_object_scores, num_anchors, anchor_scale = anchors_func.get_anchors_mode_parameters(anchors_mode, use_object_scores, num_anchors, anchor_scale)

    if isinstance(features_pick[0], str):
        fpn_features = [backbone.get_layer(layer_name) for layer_name in features_pick]
    else:
        features = model_surgery.get_pyramide_feature_layers(backbone)
        fpn_features = [features[id] for id in features_pick]
    feature_names, fpn_features = model_surgery.align_pyramide_feature_output_by_image_data_format(fpn_features)
    print(">>>> features:", {ii: jj.shape for ii, jj in zip(feature_names, fpn_features)})
    print(">>>> num_anchors:", num_anchors)

    # Build additional input features that are not from backbone.
    for id in range(additional_features):
        cur_name = "p{}_p{}_".format(id + 5, id + 6)
        additional_feature = align_feature_channel(fpn_features[-1], num_channels, name=cur_name)
        additional_feature = layers.MaxPool2D(pool_size=3, strides=2, padding="SAME", name=cur_name + "max_down")(additional_feature)
        fpn_features.append(additional_feature)

    # Bi-FPN
    for id in range(fpn_depth):
        fpn_features = bi_fpn(fpn_features, num_channels, use_weighted_sum, use_sep_conv, activation=activation, name="biFPN_{}_".format(id + 1))

    # Outputs
    bboxes_features = det_header_pre(fpn_features, num_channels, head_depth, use_sep_conv, activation=activation, name="regressor_")
    bboxes_out = det_header_post(bboxes_features, 4, num_anchors, bias_init="zeros", use_sep_conv=use_sep_conv, head_activation=None, name="regressor_")
    if use_object_scores:
        bias_init = initializers.constant(-math.log((1 - 0.01) / 0.01))
        object_out = det_header_post(bboxes_features, 1, num_anchors, bias_init, use_sep_conv, head_activation=classifier_activation, name="object_")

    if num_classes > 0:
        bias_init = initializers.constant(-math.log((1 - 0.01) / 0.01))
        class_features = det_header_pre(fpn_features, num_channels, head_depth, use_sep_conv, activation=activation, name="classifier_")
        class_out = det_header_post(class_features, num_classes, num_anchors, bias_init, use_sep_conv, classifier_activation, name="classifier_")
        if use_object_scores:
            outputs = functional.concat([bboxes_out, class_out, object_out], axis=-1)
        else:
            outputs = functional.concat([bboxes_out, class_out], axis=-1)
    else:
        outputs = functional.concat([bboxes_out, object_out], axis=-1) if use_object_scores else bboxes_out
    outputs = layers.Activation("linear", dtype="float32", name="outputs_fp32")(outputs)

    model_name = model_name or backbone.name + "_det"
    model = models.Model(inputs=backbone.inputs[0], outputs=outputs, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "efficientdet", pretrained)

    # For prediction
    # AA = {"aspect_ratios": anchor_aspect_ratios, "num_scales": anchor_num_scales, "anchor_scale": anchor_scale, "grid_zero_start": anchor_grid_zero_start}
    pyramid_levels = [pyramid_levels_min, pyramid_levels_min + len(features_pick) + additional_features - 1]  # -> [3, 7]
    post_process = eval_func.DecodePredictions(backbone.input_shape[1:], pyramid_levels, anchors_mode, use_object_scores, anchor_scale)
    add_pre_post_process(model, rescale_mode=rescale_mode, post_process=post_process)
    # model.backbone = backbone
    return model


def EfficientDetD0(input_shape=(512, 512, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="swish", pretrained="coco", **kwargs):
    if backbone is None:
        backbone_kwargs = {} if pretrained is None else {"pretrained": None}  # Load EfficientNet weights if EfficientDet pretrained not specified
        backbone = efficientnet.EfficientNetV1B0(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation, **backbone_kwargs)
        del backbone_kwargs
        features_pick = ["stack_2_block1_output", "stack_4_block2_output", "stack_6_block0_output"]
    model_name = kwargs.pop("model_name", "efficientdet_d0")
    return EfficientDet(**locals(), fpn_depth=3, head_depth=3, num_channels=64, **kwargs)


def EfficientDetD1(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="swish", pretrained="coco", **kwargs):
    if backbone is None:
        backbone_kwargs = {} if pretrained is None else {"pretrained": None}  # Load EfficientNet weights if EfficientDet pretrained not specified
        backbone = efficientnet.EfficientNetV1B1(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation, **backbone_kwargs)
        del backbone_kwargs
        features_pick = ["stack_2_block2_output", "stack_4_block3_output", "stack_6_block1_output"]
    model_name = kwargs.pop("model_name", "efficientdet_d1")
    return EfficientDet(**locals(), fpn_depth=4, head_depth=3, num_channels=88, **kwargs)


def EfficientDetD2(input_shape=(768, 768, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="swish", pretrained="coco", **kwargs):
    if backbone is None:
        backbone_kwargs = {} if pretrained is None else {"pretrained": None}  # Load EfficientNet weights if EfficientDet pretrained not specified
        backbone = efficientnet.EfficientNetV1B2(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation, **backbone_kwargs)
        del backbone_kwargs
        features_pick = ["stack_2_block2_output", "stack_4_block3_output", "stack_6_block1_output"]
    model_name = kwargs.pop("model_name", "efficientdet_d2")
    return EfficientDet(**locals(), fpn_depth=5, head_depth=3, num_channels=112, **kwargs)


def EfficientDetD3(input_shape=(896, 896, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="swish", pretrained="coco", **kwargs):
    if backbone is None:
        backbone_kwargs = {} if pretrained is None else {"pretrained": None}  # Load EfficientNet weights if EfficientDet pretrained not specified
        backbone = efficientnet.EfficientNetV1B3(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation, **backbone_kwargs)
        del backbone_kwargs
        features_pick = ["stack_2_block2_output", "stack_4_block4_output", "stack_6_block1_output"]
    model_name = kwargs.pop("model_name", "efficientdet_d3")
    return EfficientDet(**locals(), fpn_depth=6, head_depth=4, num_channels=160, **kwargs)


def EfficientDetD4(input_shape=(1024, 1024, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="swish", pretrained="coco", **kwargs):
    if backbone is None:
        backbone_kwargs = {} if pretrained is None else {"pretrained": None}  # Load EfficientNet weights if EfficientDet pretrained not specified
        backbone = efficientnet.EfficientNetV1B4(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation, **backbone_kwargs)
        del backbone_kwargs
        features_pick = ["stack_2_block3_output", "stack_4_block5_output", "stack_6_block1_output"]
    model_name = kwargs.pop("model_name", "efficientdet_d4")
    return EfficientDet(**locals(), fpn_depth=7, head_depth=4, num_channels=224, **kwargs)


def EfficientDetD5(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="swish", pretrained="coco", **kwargs):
    if backbone is None:
        backbone_kwargs = {} if pretrained is None else {"pretrained": None}  # Load EfficientNet weights if EfficientDet pretrained not specified
        backbone = efficientnet.EfficientNetV1B5(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation, **backbone_kwargs)
        del backbone_kwargs
        features_pick = ["stack_2_block4_output", "stack_4_block6_output", "stack_6_block2_output"]
    model_name = kwargs.pop("model_name", "efficientdet_d5")
    return EfficientDet(**locals(), fpn_depth=7, head_depth=4, num_channels=288, **kwargs)


def EfficientDetD6(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="swish", pretrained="coco", **kwargs):
    if backbone is None:
        backbone_kwargs = {} if pretrained is None else {"pretrained": None}  # Load EfficientNet weights if EfficientDet pretrained not specified
        backbone = efficientnet.EfficientNetV1B6(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation, **backbone_kwargs)
        del backbone_kwargs
        features_pick = ["stack_2_block5_output", "stack_4_block7_output", "stack_6_block2_output"]
    model_name = kwargs.pop("model_name", "efficientdet_d6")
    return EfficientDet(**locals(), fpn_depth=8, head_depth=5, num_channels=384, use_weighted_sum=False, **kwargs)


def EfficientDetD7(input_shape=(1536, 1536, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="swish", pretrained="coco", **kwargs):
    if backbone is None:
        backbone_kwargs = {} if pretrained is None else {"pretrained": None}  # Load EfficientNet weights if EfficientDet pretrained not specified
        backbone = efficientnet.EfficientNetV1B6(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation, **backbone_kwargs)
        del backbone_kwargs
        features_pick = ["stack_2_block5_output", "stack_4_block7_output", "stack_6_block2_output"]
    anchor_scale = kwargs.pop("anchor_scale", 5)
    model_name = kwargs.pop("model_name", "efficientdet_d7")
    return EfficientDet(**locals(), fpn_depth=8, head_depth=5, num_channels=384, use_weighted_sum=False, **kwargs)


def EfficientDetD7X(input_shape=(1536, 1536, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="swish", pretrained="coco", **kwargs):
    if backbone is None:
        backbone_kwargs = {} if pretrained is None else {"pretrained": None}  # Load EfficientNet weights if EfficientDet pretrained not specified
        backbone = efficientnet.EfficientNetV1B7(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation, **backbone_kwargs)
        del backbone_kwargs
        features_pick = ["stack_2_block6_output", "stack_4_block9_output", "stack_6_block3_output"]
    additional_features = 3
    model_name = kwargs.pop("model_name", "efficientdet_d7x")
    return EfficientDet(**locals(), fpn_depth=8, head_depth=5, num_channels=384, use_weighted_sum=False, **kwargs)


""" EfficientDetLite models """


def EfficientDetLite0(input_shape=(320, 320, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="relu6", pretrained="coco", **kwargs):
    if backbone is None:
        backbone = efficientnet.EfficientNetV1Lite0(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation)
        features_pick = ["stack_2_block1_output", "stack_4_block2_output", "stack_6_block0_output"]
    model_name = kwargs.pop("model_name", "efficientdet_lite0")
    anchor_scale = kwargs.pop("anchor_scale", 3)
    return EfficientDet(**locals(), fpn_depth=3, head_depth=3, num_channels=64, use_weighted_sum=False, rescale_mode="tf", **kwargs)


def EfficientDetLite1(input_shape=(384, 384, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="relu6", pretrained="coco", **kwargs):
    if backbone is None:
        backbone = efficientnet.EfficientNetV1Lite1(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation)
        features_pick = ["stack_2_block2_output", "stack_4_block3_output", "stack_6_block0_output"]
    model_name = kwargs.pop("model_name", "efficientdet_lite1")
    anchor_scale = kwargs.pop("anchor_scale", 3)
    return EfficientDet(**locals(), fpn_depth=4, head_depth=3, num_channels=88, use_weighted_sum=False, rescale_mode="tf", **kwargs)


def EfficientDetLite2(input_shape=(448, 448, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="relu6", pretrained="coco", **kwargs):
    if backbone is None:
        backbone = efficientnet.EfficientNetV1Lite2(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation)
        features_pick = ["stack_2_block2_output", "stack_4_block3_output", "stack_6_block0_output"]
    model_name = kwargs.pop("model_name", "efficientdet_lite2")
    anchor_scale = kwargs.pop("anchor_scale", 3)
    return EfficientDet(**locals(), fpn_depth=5, head_depth=3, num_channels=112, use_weighted_sum=False, rescale_mode="tf", **kwargs)


def EfficientDetLite3(input_shape=(512, 512, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="relu6", pretrained="coco", **kwargs):
    if backbone is None:
        backbone = efficientnet.EfficientNetV1Lite3(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation)
        features_pick = ["stack_2_block2_output", "stack_4_block4_output", "stack_6_block0_output"]
    model_name = kwargs.pop("model_name", "efficientdet_lite3")
    anchor_scale = kwargs.pop("anchor_scale", 4)
    return EfficientDet(**locals(), fpn_depth=6, head_depth=4, num_channels=160, use_weighted_sum=False, rescale_mode="tf", **kwargs)


def EfficientDetLite3X(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="relu6", pretrained="coco", **kwargs):
    if backbone is None:
        backbone = efficientnet.EfficientNetV1Lite3(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation)
        features_pick = ["stack_2_block2_output", "stack_4_block4_output", "stack_6_block0_output"]
    model_name = kwargs.pop("model_name", "efficientdet_lite3x")
    anchor_scale = kwargs.pop("anchor_scale", 3)
    return EfficientDet(**locals(), fpn_depth=6, head_depth=4, num_channels=200, use_weighted_sum=False, rescale_mode="tf", **kwargs)


def EfficientDetLite4(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=90, backbone=None, activation="relu6", pretrained="coco", **kwargs):
    if backbone is None:
        backbone = efficientnet.EfficientNetV1Lite4(input_shape=input_shape, num_classes=0, output_conv_filter=0, activation=activation)
        features_pick = ["stack_2_block3_output", "stack_4_block5_output", "stack_6_block0_output"]
    model_name = kwargs.pop("model_name", "efficientdet_lite4")
    anchor_scale = kwargs.pop("anchor_scale", 4)
    return EfficientDet(**locals(), fpn_depth=7, head_depth=4, num_channels=224, use_weighted_sum=False, rescale_mode="tf", **kwargs)



from keras_cv_attention_models.aotnet import AotNet
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "regnety_040": {"imagenet": "35bde7c2f9e391c20d4aa20160902961"},
    "regnety_064": {"imagenet": "1b096c91cdcf5b3ea82900468cda829b"},
    "regnety_080": {"imagenet": "651b4e500556593695258dee604eecba"},
    "regnety_160": {"imagenet": "7a039bfb1e006571991a19b5e3fbd0d1"},
    "regnety_320": {"imagenet": "20fbbf6de39635b89ae284ac0574c4cf"},
    "regnetz_b16": {"imagenet": "9f03b9e885c6918db54374b74f355eb9"},
    "regnetz_c16": {"imagenet": "cfa0cf70aa724d8e79633641208d7d39"},
    "regnetz_c16_evo": {"imagenet": "cf04ad48e51d652d901b03e195dace21"},
    "regnetz_d32": {"imagenet": "b76e7c68a309a3697ffb5f30c0aeeea6"},
    "regnetz_d8": {"imagenet": "1fa2de5606f33b503cedda4689b8bddd"},
    "regnetz_d8_evo": {"imagenet": "97b8bcf98755868765b3bc62700bbb39"},
    "regnetz_e8": {"imagenet": "7d0a85b287ad041e0446623c7070fa8f"},
}


def RegNetY(num_blocks, out_channels, input_shape=(224, 224, 3), hidden_channel_ratio=1, stem_width=32, se_ratio=0.25, pretrained="imagenet", **kwargs):
    strides = [2, 2, 2, 2]
    stem_type = "kernel_3x3"
    stem_downsample = False
    # se_ratio using input_channel
    # se_ratio = [[se_ratio * (stem_width if id == 0 else out_channels[id - 1]) / out_channels[id]] + [se_ratio] * bb for id, bb in enumerate(num_blocks)]
    se_ratio = [
        [se_ratio * stem_width / out_channels[0]] + [se_ratio] * num_blocks[0],
        [se_ratio * out_channels[0] / out_channels[1]] + [se_ratio] * num_blocks[1],
        [se_ratio * out_channels[1] / out_channels[2]] + [se_ratio] * num_blocks[2],
        [se_ratio * out_channels[2] / out_channels[3]] + [se_ratio] * num_blocks[3],
    ]
    attn_params = {"se_divisor": 1}
    kwargs.pop("kwargs", None)
    model = AotNet(**locals(), **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnet_family", pretrained=pretrained)
    return model


def RegNetY032(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 5, 13, 1]
    out_channels = [72, 216, 576, 1512]
    group_size = 24
    model = RegNetY(**locals(), model_name="regnety_032", **kwargs)
    return model


def RegNetY040(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 6, 12, 2]
    out_channels = [128, 192, 512, 1088]
    group_size = 64
    model = RegNetY(**locals(), model_name="regnety_040", **kwargs)
    return model


def RegNetY064(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 7, 14, 2]
    out_channels = [144, 288, 576, 1296]
    group_size = 72
    model = RegNetY(**locals(), model_name="regnety_064", **kwargs)
    return model


def RegNetY080(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 4, 10, 1]
    out_channels = [168, 448, 896, 2016]
    group_size = 56
    model = RegNetY(**locals(), model_name="regnety_080", **kwargs)
    return model


def RegNetY160(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 4, 11, 1]
    out_channels = [224, 448, 1232, 3024]
    group_size = 112
    model = RegNetY(**locals(), model_name="regnety_160", **kwargs)
    return model


def RegNetY320(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 5, 12, 1]
    out_channels = [232, 696, 1392, 3712]
    group_size = 232
    model = RegNetY(**locals(), model_name="regnety_320", **kwargs)
    return model


def RegNetZB16(input_shape=(224, 224, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 6, 12, 2]
    strides = [2, 2, 2, 2]
    out_channels = [48, 96, 192, 288]
    # timm bottle_in=True mode, the first ratio in each stack is `ratio * previous_channel`
    hidden_channel_ratio = [[32 * 3 / 48, 3], [1.5] + [3] * 5, [1.5] + [3] * 11, [192 * 3 / 288, 3]]
    use_block_output_activation = False  # timm linear_out=True mode
    stem_type = "kernel_3x3"
    stem_width = 32
    stem_downsample = False
    se_ratio = 0.25
    attn_params = {"activation": "relu"}
    group_size = 16
    shortcut_type = None
    output_num_features = 1536
    kwargs.pop("kwargs", None)
    model = AotNet(**locals(), model_name=kwargs.pop("model_name", "regnetz_b16"), **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnet_family", pretrained=pretrained)
    return model


def RegNetZC16(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 6, 12, 2]
    strides = [2, 2, 2, 2]
    out_channels = [48, 96, 192, 288]
    # timm bottle_in=True mode, the first ratio in each stack is `ratio * previous_channel`
    hidden_channel_ratio = [[32 * 4 / 48, 4], [2] + [4] * 5, [2] + [4] * 11, [192 * 4 / 288, 4]]
    use_block_output_activation = False  # timm linear_out=True mode
    stem_type = "kernel_3x3"
    stem_width = 32
    stem_downsample = False
    se_ratio = 0.25
    attn_params = {"activation": "relu"}
    group_size = 16
    shortcut_type = None
    output_num_features = 1536
    kwargs.pop("kwargs", None)
    model = AotNet(**locals(), model_name=kwargs.pop("model_name", "regnetz_c16"), **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnet_family", pretrained=pretrained)
    return model


def RegNetZC16_EVO(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    use_evo_norm = True
    evo_norm_group_size = 16
    bn_epsilon = 1e-3
    return RegNetZC16(**locals(), model_name="regnetz_c16_evo", **kwargs)


def RegNetZD32(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 6, 12, 3]
    strides = [1, 2, 2, 2]
    out_channels = [64, 128, 256, 384]
    # timm bottle_in=True mode, the first ratio in each stack is `ratio * previous_channel`
    hidden_channel_ratio = [[64 * 4 / 64, 4, 4], [2] + [4] * 5, [2] + [4] * 11, [256 * 4 / 384, 4, 4]]
    use_block_output_activation = False  # timm linear_out=True mode
    stem_type = "tiered"
    stem_last_strides = 2
    stem_width = 64
    stem_downsample = False
    se_ratio = 0.25
    attn_params = {"activation": "relu"}
    group_size = 32
    shortcut_type = None
    output_num_features = 1792
    kwargs.pop("kwargs", None)
    model = AotNet(**locals(), model_name=kwargs.pop("model_name", "regnetz_d32"), **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnet_family", pretrained=pretrained)
    return model


def RegNetZD8(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 6, 12, 3]
    strides = [1, 2, 2, 2]
    out_channels = [64, 128, 256, 384]
    # timm bottle_in=True mode, the first ratio in each stack is `ratio * previous_channel`
    hidden_channel_ratio = [[64 * 4 / 64, 4, 4], [64 * 4 / 128] + [4] * 5, [128 * 4 / 256] + [4] * 11, [256 * 4 / 384, 4, 4]]
    use_block_output_activation = False  # timm linear_out=True mode
    stem_type = kwargs.pop("stem_type", "tiered")  # support RegNetZD8_EVO
    stem_last_strides = 2
    stem_width = 64
    stem_downsample = False
    se_ratio = 0.25
    attn_params = {"activation": "relu"}
    group_size = 8
    shortcut_type = None
    output_num_features = 1792
    kwargs.pop("kwargs", None)
    model = AotNet(**locals(), model_name=kwargs.pop("model_name", "regnetz_d8"), **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnet_family", pretrained=pretrained)
    return model


def RegNetZD8_EVO(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    stem_type = "deep2"
    use_evo_norm = True
    evo_norm_group_size = 16
    bn_epsilon = 1e-3
    return RegNetZD8(**locals(), model_name="regnetz_d8_evo", **kwargs)


def RegNetZE8(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 8, 16, 3]
    strides = [1, 2, 2, 2]
    out_channels = [96, 192, 384, 512]
    # timm bottle_in=True mode, the first ratio in each stack is `ratio * previous_channel`
    hidden_channel_ratio = [[64 * 4 / 96, 4, 4], [96 * 4 / 192] + [4] * 7, [192 * 4 / 384] + [4] * 15, [384 * 4 / 512, 4, 4]]
    use_block_output_activation = False  # timm linear_out=True mode
    stem_type = "tiered"
    stem_last_strides = 2
    stem_width = 64
    stem_downsample = False
    se_ratio = 0.25
    attn_params = {"activation": "relu"}
    group_size = 8
    shortcut_type = None
    output_num_features = 2048
    kwargs.pop("kwargs", None)
    model = AotNet(**locals(), model_name=kwargs.pop("model_name", "regnetz_e8"), **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnet_family", pretrained=pretrained)
    return model



from keras_cv_attention_models.halonet.halonet import (
    halo_attention,
    HaloNet,
    HaloNetH0,
    HaloNetH1,
    HaloNetH2,
    HaloNetH3,
    HaloNetH4,
    HaloNetH5,
    HaloNetH6,
    HaloNetH7,
    HaloNet26T,
    HaloNet50T,
    HaloNetSE33T,
    HaloNextECA26T,
    HaloRegNetZB,
    HaloBotNet50T,
)

__head_doc__ = """
Keras implementation of [Github lucidrains/halonet-pytorch](https://github.com/lucidrains/halonet-pytorch).
Paper [PDF 2103.12731 Scaling Local Self-Attention for Parameter Efficient Visual Backbones](https://arxiv.org/pdf/2103.12731.pdf).
"""

__tail_doc__ = """  out_channels: output channels for each stack, default `[64, 128, 256, 512]`.
  strides: strides for each stack, default `[1, 2, 2, 2]`.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `softmax`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.
      No pretrained available for `H` models.
  **kwargs: other parameters from `AotNet` if not conflict.

Returns:
    A `keras.Model` instance.
"""

HaloNet.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  halo_block_size: block_size for halo attention layers.
  halo_halo_size: halo_size for halo attention layers.
  halo_expansion: filter expansion for halo attention output channel.
  num_heads: number of heads in each stack. Can be a number or list of number like `[4, 8, 8, 8]`.
  expansion: filter expansion for each block output channel.
  output_num_features: none `0` value to add another `conv2d + bn + activation` layers before `GlobalAveragePooling2D`.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model          | Params | FLOPs   | Input | Top1 Acc |
  | -------------- | ------ | ------- | ----- | -------- |
  | HaloNetH0      | 5.5M   | 2.40G   | 256   | 77.9     |
  | HaloNetH1      | 8.1M   | 3.04G   | 256   | 79.9     |
  | HaloNetH2      | 9.4M   | 3.37G   | 256   | 80.4     |
  | HaloNetH3      | 11.8M  | 6.30G   | 320   | 81.9     |
  | HaloNetH4      | 19.1M  | 12.17G  | 384   | 83.3     |
  | - 21k          | 19.1M  | 12.17G  | 384   | 85.5     |
  | HaloNetH5      | 30.7M  | 32.61G  | 448   | 84.0     |
  | HaloNetH6      | 43.4M  | 53.20G  | 512   | 84.4     |
  | HaloNetH7      | 67.4M  | 119.64G | 600   | 84.9     |
  | HaloNextECA26T | 10.7M  | 2.43G   | 256   | 79.50    |
  | HaloNet26T     | 12.5M  | 3.18G   | 256   | 79.13    |
  | HaloNetSE33T   | 13.7M  | 3.55G   | 256   | 80.99    |
  | HaloRegNetZB   | 11.68M | 1.97G   | 224   | 81.042   |
  | HaloNet50T     | 22.7M  | 5.29G   | 256   | 81.70    |
  | HaloBotNet50T  | 22.6M  | 5.02G   | 256   | 82.0     |

"""

HaloNetH0.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

HaloNetH1.__doc__ = HaloNetH0.__doc__
HaloNetH2.__doc__ = HaloNetH0.__doc__
HaloNetH3.__doc__ = HaloNetH0.__doc__
HaloNetH4.__doc__ = HaloNetH0.__doc__
HaloNetH5.__doc__ = HaloNetH0.__doc__
HaloNetH6.__doc__ = HaloNetH0.__doc__
HaloNetH7.__doc__ = HaloNetH0.__doc__

HaloNet26T.__doc__ = __head_doc__ + """Model weights are reloaded from timm [Github rwightman/pytorch-image-models](https://github.com/rwightman/pytorch-image-models).

Args:
""" + __tail_doc__
HaloNetSE33T.__doc__ = HaloNet26T.__doc__
HaloNextECA26T.__doc__ = HaloNet26T.__doc__
HaloNet50T.__doc__ = HaloNet26T.__doc__
HaloRegNetZB.__doc__ = HaloNet26T.__doc__
HaloBotNet50T.__doc__ = HaloNet26T.__doc__

halo_attention.__doc__ = __head_doc__ + """
Halo Attention. Defined as function, not layer.
Extract patches with a `kernel_size` from `key_value` as an enlarged attention area.
Also adds `RelativePositionalEmbedding` to `attention_scores`.

Args:
  inputs: input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query and key. Default `0` for `key_dim = inputs.shape[-1] // num_heads`
  block_size: works like `kernel_size` from `Conv2D`, extracting input patches as `query`.
  halo_size: expansion to `block_size`, extracting input patches as `key` and `value`.
  strides: downsample strides for `query`.
  out_shape: The expected shape of an output tensor. If not specified, projects back to the input dim.
  out_weight: Boolean, whether use an ouput dense.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  attn_dropout: Dropout probability for attention.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.halo_attention(inputs, num_heads=4)
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 256])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'conv2d_1/kernel:0': TensorShape([1, 1, 256, 512]),
#  'conv2d/kernel:0': TensorShape([1, 1, 256, 256]),
#  'relative_positional_embedding/r_height:0': TensorShape([64, 7]),
#  'relative_positional_embedding/r_width:0': TensorShape([64, 7]),
#  'dense/kernel:0': TensorShape([256, 256])}
"""



import os

is_tensorflow_backend = not "torch" in os.getenv("KECAM_BACKEND", "tensorflow").lower()

if is_tensorflow_backend:
    import tensorflow as tf
    from tensorflow.keras import layers, models, initializers, callbacks
    from tensorflow.keras.utils import register_keras_serializable, get_file
    from keras_cv_attention_models import tf_functional as functional
else:
    from keras_cv_attention_models.pytorch_backend import layers, models, functional, initializers, callbacks
    from keras_cv_attention_models.pytorch_backend.utils import register_keras_serializable, get_file

    print(">>>> Using PyTorch backend")


def backend():
    if is_tensorflow_backend:
        return tf.keras.backend.backend()
    else:
        return "pytorch"


def image_data_format():
    if is_tensorflow_backend:
        return tf.keras.backend.image_data_format()
    else:
        return "channels_first"


__is_channels_last__ = image_data_format() == "channels_last"


def is_channels_last():
    return __is_channels_last__


def align_input_shape_by_image_data_format(input_shape):
    """Regard input_shape as force using original shape if len(input_shape) == 4,
    else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format,
    or if dynamic shape liek `(None, None, 3)`, will regard the known shape 3 as channel dimentsion.

    Examples:
    >>> os.environ['KECAM_BACKEND'] = "torch"
    >>> from keras_cv_attention_models import backend
    >>> print(f"{backend.image_data_format() = }")
    >>> # backend.image_data_format() = 'channels_first'
    >>> print(backend.align_input_shape_by_image_data_format([224, 224, 3]))
    >>> # [3, 224, 224]
    >>> print(backend.align_input_shape_by_image_data_format([3, 224, 224]))
    >>> # [3, 224, 224]
    >>> print(backend.align_input_shape_by_image_data_format([None, None, 3]))
    >>> # [3, None, None]
    >>> print(backend.align_input_shape_by_image_data_format([None, 224, 224, 3]))
    >>> # [224, 224, 3]
    """
    # channel_axis = if image_data_format() == "channels_last" else
    if len(input_shape) == 4:  # Regard this as force using original shape
        return input_shape[1:]

    # Assume channel dimention is the one with min value in input_shape
    if None in input_shape:
        orign_channel_axis, channel_dim = min(enumerate(input_shape), key=lambda xx: xx[1] is None)
    else:
        orign_channel_axis, channel_dim = min(enumerate(input_shape), key=lambda xx: xx[1])
    orign_block_shape = [dim for axis, dim in enumerate(input_shape) if axis != orign_channel_axis]
    aligned = [*orign_block_shape, channel_dim] if image_data_format() == "channels_last" else [channel_dim, *orign_block_shape]
    if aligned[1] != input_shape[1] or aligned[-1] != input_shape[-1]:
        print(">>>> Aligned input_shape:", aligned)
    return aligned


def in_train_phase(train_phase, eval_phase, training=None):
    if is_tensorflow_backend:
        return tf.keras.backend.in_train_phase(train_phase, eval_phase, training=training)
    else:
        return train_phase if training else eval_phase  # [TODO]


def numpy_image_resize(inputs, target_shape, method="bilinear", is_source_channels_last=True):
    ndims = len(inputs.shape)
    if ndims < 2 or ndims > 4:
        raise ValueError("inputs with shape={}, ndims={} not supported, ndims must in [2, 4]".format(ipnuts.shape, ndims))

    if is_source_channels_last:
        inputs = inputs if ndims == 4 else (inputs[None] if ndims == 3 else inputs[None, :, :, None])
        inputs = inputs if image_data_format() == "channels_last" else inputs.transpose([0, 3, 1, 2])

        inputs = functional.resize(inputs, target_shape, method=method)
        inputs = inputs.detach().numpy() if hasattr(inputs, "detach") else inputs.numpy()

        inputs = inputs if image_data_format() == "channels_last" else inputs.transpose([0, 2, 3, 1])
        inputs = inputs if ndims == 4 else (inputs[0] if ndims == 3 else inputs[0, :, :, 0])
    else:
        inputs = inputs if ndims == 4 else (inputs[None] if ndims == 3 else inputs[None, None, :, :])
        inputs = inputs.transpose([0, 2, 3, 1]) if image_data_format() == "channels_last" else inputs

        inputs = functional.resize(inputs, target_shape, method=method)
        inputs = inputs.detach().numpy() if hasattr(inputs, "detach") else inputs.numpy()

        inputs = inputs.transpose([0, 3, 1, 2]) if image_data_format() == "channels_last" else inputs
        inputs = inputs if ndims == 4 else (inputs[0] if ndims == 3 else inputs[0, 0])
    return inputs



import os
import json
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K


class CosineLrScheduler(keras.callbacks.Callback):
    def __init__(self, lr_base, first_restart_step, steps_per_epoch=-1, m_mul=0.5, t_mul=2.0, lr_min=1e-5, lr_warmup=-1, warmup_steps=0, cooldown_steps=0):
        super(CosineLrScheduler, self).__init__()
        self.lr_base, self.m_mul, self.t_mul, self.lr_min, self.steps_per_epoch = lr_base, m_mul, t_mul, lr_min, steps_per_epoch
        self.first_restart_step, self.warmup_steps, self.cooldown_steps, self.lr_warmup = first_restart_step, warmup_steps, cooldown_steps, lr_warmup
        self.init_step_num, self.cur_epoch, self.is_cooldown_epoch, self.previous_cooldown_steps = 0, 0, False, 0
        self.is_built = False
        if steps_per_epoch != -1:
            self.build(steps_per_epoch)

    def build(self, steps_per_epoch=-1):
        # print(">>>> steps_per_epoch:", steps_per_epoch)
        if steps_per_epoch != -1:
            self.steps_per_epoch = steps_per_epoch

        first_restart_batch_step = self.first_restart_step * self.steps_per_epoch
        alpha = self.lr_min / self.lr_base
        if self.lr_min == self.lr_base * self.m_mul:  # Without restart
            self.schedule = keras.experimental.CosineDecay(self.lr_base, first_restart_batch_step, alpha=alpha)
            self.cooldown_steps_start, self.cooldown_steps_end = np.array([]), np.array([])
        else:
            self.schedule = keras.experimental.CosineDecayRestarts(self.lr_base, first_restart_batch_step, t_mul=self.t_mul, m_mul=self.m_mul, alpha=alpha)
            aa = [first_restart_batch_step / self.steps_per_epoch * (self.t_mul**ii) for ii in range(5)]
            self.cooldown_steps_start = np.array([int(sum(aa[:ii]) + self.cooldown_steps * (ii - 1)) for ii in range(1, 5)])
            self.cooldown_steps_end = np.array([ii + self.cooldown_steps for ii in self.cooldown_steps_start])

        if self.warmup_steps != 0:
            self.warmup_batch_steps = self.warmup_steps * self.steps_per_epoch
            self.lr_warmup = self.lr_warmup if self.lr_warmup > 0 else self.lr_min
            self.warmup_lr_func = lambda ii: self.lr_warmup + (self.lr_base - self.lr_warmup) * ii / self.warmup_batch_steps
        else:
            self.warmup_batch_steps = 0
        self.is_built = True

    def on_epoch_begin(self, cur_epoch, logs=None):
        if not self.is_built:
            self.build()

        self.init_step_num = int(self.steps_per_epoch * cur_epoch)
        self.cur_epoch = cur_epoch

        if self.cooldown_steps_end.shape[0] != 0:
            cooldown_end_pos = (self.cooldown_steps_end > cur_epoch).argmax()
            self.previous_cooldown_steps = self.cooldown_steps * cooldown_end_pos * self.steps_per_epoch
            if cur_epoch >= self.cooldown_steps_end[cooldown_end_pos] - self.cooldown_steps:
                self.is_cooldown_epoch = True
            else:
                self.is_cooldown_epoch = False

    def on_train_batch_begin(self, iterNum, logs=None):
        global_iterNum = iterNum + self.init_step_num
        if global_iterNum < self.warmup_batch_steps:
            lr = self.warmup_lr_func(global_iterNum)
        elif self.is_cooldown_epoch:
            lr = self.lr_min  # cooldown
        else:
            # lr = self.schedule(global_iterNum - self.warmup_batch_steps - self.previous_cooldown_steps)
            lr = self.schedule(global_iterNum - self.previous_cooldown_steps)

        if self.model is not None:
            K.set_value(self.model.optimizer.lr, lr)
        if iterNum == 0:
            print("\nLearning rate for iter {} is {}, global_iterNum is {}".format(self.cur_epoch + 1, lr, global_iterNum))
        return lr


class CosineLrSchedulerEpoch(keras.callbacks.Callback):
    def __init__(self, lr_base, first_restart_step, m_mul=0.5, t_mul=2.0, lr_min=1e-6, lr_warmup=-1, warmup_steps=0, cooldown_steps=0):
        super(CosineLrSchedulerEpoch, self).__init__()
        self.warmup_steps, self.cooldown_steps, self.lr_min = warmup_steps, cooldown_steps, lr_min

        if lr_min == lr_base * m_mul:
            self.schedule = keras.experimental.CosineDecay(lr_base, first_restart_step, alpha=lr_min / lr_base)
            self.cooldown_steps_start, self.cooldown_steps_end = np.array([]), np.array([])
        else:
            self.schedule = keras.experimental.CosineDecayRestarts(lr_base, first_restart_step, t_mul=t_mul, m_mul=m_mul, alpha=lr_min / lr_base)
            aa = [first_restart_step * (t_mul**ii) for ii in range(5)]
            self.cooldown_steps_start = np.array([int(sum(aa[:ii]) + cooldown_steps * (ii - 1)) for ii in range(1, 5)])
            self.cooldown_steps_end = np.array([ii + cooldown_steps for ii in self.cooldown_steps_start])

        if warmup_steps != 0:
            self.lr_warmup = lr_warmup if lr_warmup > 0 else lr_min
            self.warmup_lr_func = lambda ii: self.lr_warmup + (lr_base - self.lr_warmup) * ii / warmup_steps

    def on_epoch_begin(self, epoch, logs=None):
        if epoch < self.warmup_steps:
            lr = self.warmup_lr_func(epoch)
        elif self.cooldown_steps_end.shape[0] != 0:
            cooldown_end_pos = (self.cooldown_steps_end > epoch).argmax()
            if epoch >= self.cooldown_steps_end[cooldown_end_pos] - self.cooldown_steps:
                lr = self.lr_min  # cooldown
            else:
                # lr = self.schedule(epoch - self.warmup_steps)
                lr = self.schedule(epoch - self.cooldown_steps * cooldown_end_pos)
        else:
            lr = self.schedule(epoch)

        if self.model is not None:
            K.set_value(self.model.optimizer.lr, lr)

        print("\nLearning rate for iter {} is {}".format(epoch + 1, lr))
        return lr


def constant_scheduler(epoch, lr_base, lr_decay_steps, decay_rate=0.1, warmup_steps=0):
    if epoch < warmup_steps:
        lr = lr_base * (epoch + 1) / (warmup_steps + 1)
    else:
        # epoch -= warmup_steps
        lr = lr_base * decay_rate ** np.sum(epoch >= np.array(lr_decay_steps))
    print("\nLearning rate for iter {} is {}".format(epoch + 1, lr))
    return lr


def exp_scheduler(epoch, lr_base=0.1, decay_step=1, decay_rate=0.9, lr_min=0, warmup_steps=0):
    if epoch < warmup_steps:
        lr = (lr_base - lr_min) * (epoch + 1) / (warmup_steps + 1)
    else:
        # epoch -= warmup_steps
        lr = lr_base * decay_rate ** (epoch / decay_step)
        lr = lr if lr > lr_min else lr_min
    # print("Learning rate for iter {} is {}".format(epoch + 1, lr))
    return lr


class OptimizerWeightDecay(keras.callbacks.Callback):
    def __init__(self, lr_base, wd_base, is_lr_on_batch=False):
        super(OptimizerWeightDecay, self).__init__()
        self.wd_m = wd_base / lr_base
        self.lr_base, self.wd_base = lr_base, wd_base
        # self.model.optimizer.weight_decay = lambda: wd_m * self.model.optimizer.lr
        self.is_lr_on_batch = is_lr_on_batch
        if is_lr_on_batch:
            self.on_train_batch_begin = self.__update_wd__
        else:
            self.on_epoch_begin = self.__update_wd__

    def __update_wd__(self, step, log=None):
        if self.model is not None:
            wd = self.wd_m * K.get_value(self.model.optimizer.lr)
            # wd = self.wd_base * K.get_value(self.model.optimizer.lr)
            K.set_value(self.model.optimizer.weight_decay, wd)
        # wd = self.model.optimizer.weight_decay
        if not self.is_lr_on_batch or step == 0:
            print("Weight decay is {}".format(wd))


class MyHistory(keras.callbacks.Callback):
    def __init__(self, initial_file=None):
        super(MyHistory, self).__init__()
        if initial_file and os.path.exists(initial_file):
            with open(initial_file, "r") as ff:
                self.history = json.load(ff)
        else:
            self.history = {}
        self.initial_file = initial_file

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        logs.pop("lr", None)
        lr = self.model.optimizer.lr
        if hasattr(lr, "value"):
            lr = lr.value()

        self.history.setdefault("lr", []).append(float(lr))
        for k, v in logs.items():
            k = "accuracy" if "accuracy" in k else k
            self.history.setdefault(k, []).append(float(v))

        if len(self.model.losses) != 0:  # Has regular_loss
            regular_loss = K.sum(self.model.losses).numpy()
            self.history.setdefault("regular_loss", []).append(float(regular_loss))
            self.history["loss"][-1] -= regular_loss
            if "val_loss" in self.history:
                self.history["val_loss"][-1] -= regular_loss

        if "val_ap_ar" in self.model.history.history:  # save coco val_ap_ar
            self.history.setdefault("val_ap_ar", []).append(self.model.history.history["val_ap_ar"][-1])

        if self.initial_file:
            with open(self.initial_file, "w") as ff:
                json.dump(self.history, ff)

    def print_hist(self):
        print("{")
        for kk, vv in self.history.items():
            print("  '%s': %s," % (kk, vv))
        print("}")


class MyCheckpoint(keras.callbacks.Callback):
    """Save latest and best one"""

    def __init__(self, basic_save_name, monitor="val_acc", mode="auto", save_path="checkpoints"):
        super(MyCheckpoint, self).__init__()
        self.basic_save_name, self.mode, self.save_path = basic_save_name, mode, save_path
        self.__init_monitor_strategy__(monitor)

    def __init_monitor_strategy__(self, monitor):
        self.monitor = monitor
        self.monitor_save = os.path.join(self.save_path, self.basic_save_name + "_epoch_{}_" + monitor + "_{}.h5")
        self.monitor_save_re = self.monitor_save.format("*", "*")
        self.latest_save = os.path.join(self.save_path, self.basic_save_name + "_latest.h5")
        self.is_better = (lambda cur, pre: cur <= pre) if self.mode == "min" or "loss" in monitor else (lambda cur, pre: cur >= pre)
        self.pre_best = 1e5 if self.mode == "min" or "loss" in monitor else -1e5

    def on_epoch_end(self, epoch, logs={}):
        # tf.print(">>>> Save latest to:", self.latest_save)
        if self.model is not None:
            self.model.save(self.latest_save)
        if self.monitor is not None and self.monitor not in logs:
            all_val_acc = [ii for ii in logs.keys() if "val" in ii and "acc" in ii]
            all_val_loss = [ii for ii in logs.keys() if "val" in ii and "loss" in ii]
            if len(all_val_acc) > 0:
                self.__init_monitor_strategy__(all_val_acc[0])
            elif len(all_val_loss) > 0:
                self.__init_monitor_strategy__(all_val_loss[0])
            else:
                # self.__init_monitor_strategy__("loss")
                self.monitor = None  # Not saving if not using eval dataset

        cur_monitor_val = logs.get(self.monitor, self.pre_best)
        if self.monitor is not None and self.is_better(cur_monitor_val, self.pre_best):
            self.pre_best = cur_monitor_val
            pre_monitor_saves = tf.io.gfile.glob(self.monitor_save_re)
            # tf.print(">>>> pre_monitor_saves:", pre_monitor_saves)
            if len(pre_monitor_saves) != 0:
                os.remove(pre_monitor_saves[0])
            monitor_save = self.monitor_save.format(epoch + 1, "{:.4f}".format(cur_monitor_val))
            tf.print("\n>>>> Save best to:", monitor_save)
            if self.model is not None:
                self.model.save(monitor_save)



from keras_cv_attention_models.resnet_family.resnext import ResNeXt, ResNeXt50, ResNeXt101, ResNeXt50D, ResNeXt101W, ResNeXt101W_64
from keras_cv_attention_models.resnet_family.resnet_quad import ResNetQ, ResNet51Q, ResNet61Q
from keras_cv_attention_models.resnet_family.resnet_deep import ResNetD, ResNet50D, ResNet101D, ResNet152D, ResNet200D
from keras_cv_attention_models.resnet_family.regnet import (
    RegNetY,
    RegNetY040,
    RegNetY064,
    RegNetY080,
    RegNetY160,
    RegNetY320,
    RegNetZB16,
    RegNetZC16,
    RegNetZC16_EVO,
    RegNetZD32,
    RegNetZD8,
    RegNetZD8_EVO,
    RegNetZE8
)


__resnext_head_doc__ = """
Keras implementation of [Github facebookresearch/ResNeXt](https://github.com/facebookresearch/ResNeXt).
Paper [PDF 1611.05431 Aggregated Residual Transformations for Deep Neural Networks](https://arxiv.org/pdf/1611.05431.pdf).

Pre-trained `swsl` means `Semi-Weakly Supervised ResNe*t`
    from [Github facebookresearch/semi-supervised-ImageNet1K-models](https://github.com/facebookresearch/semi-supervised-ImageNet1K-models).
    **Please note the CC-BY-NC 4.0 license on theses weights, non-commercial use only**.
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, default `(224, 224, 3)`.
    Set `(None, None, 3)` for dynamic input shape.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `softmax`.
  pretrained: value in {pretrained_list}.
      Will try to download and load pre-trained model weights if not None.
      Save path is `~/.keras/models/`.
  **kwargs: other parameters in `keras_cv_attention_models.aotnet.AotNet` if not conflict.

Returns:
    A `keras.Model` instance.
"""

ResNeXt.__doc__ = __resnext_head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: Output channel for each stack.
      `[128, 256, 512, 1024]`. for regular `32x4d` models,
      `[256, 512, 1024, 2048]` for `32x8d` wider models.
  strides: a `number` or `list`, indicates strides used in the last stack or list value for all stacks.
      If a number, it will be `[1, 2, 2, strides]`. Default `2`.
  model_name: string, model name.
""" + __tail_doc__.format(pretrained_list=[None, "imagenet", "swsl"]) + """
Model architectures:
  | Model                     | Params | FLOPs  | Input | Top1 Acc |
  | ------------------------- | ------ | ------ | ----- | -------- |
  | ResNeXt50 (32x4d)         | 25M    | 4.23G  | 224   | 79.768   |
  | - SWSL                    | 25M    | 4.23G  | 224   | 82.182   |
  | ResNeXt50D (32x4d + deep) | 25M    | 4.47G  | 224   | 79.676   |
  | ResNeXt101 (32x4d)        | 42M    | 7.97G  | 224   | 80.334   |
  | - SWSL                    | 42M    | 7.97G  | 224   | 83.230   |
  | ResNeXt101W (32x8d)       | 89M    | 16.41G | 224   | 79.308   |
  | - SWSL                    | 89M    | 16.41G | 224   | 84.284   |
  | ResNeXt101W_64 (64x4d)    | 83.46M | 15.46G | 224   | 82.46    |
"""

__resnext_default_doc__ = __resnext_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained_list=[None, "imagenet", "swsl"])

ResNeXt50.__doc__ = __resnext_default_doc__
ResNeXt101.__doc__ = __resnext_default_doc__
ResNeXt101W.__doc__ = __resnext_default_doc__
ResNeXt50D.__doc__ = __resnext_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"])
ResNeXt101W_64.__doc__ = ResNeXt50D.__doc__

__resnetd_head_doc__ = """
Github source [leondgarse/keras_cv_attention_models](https://github.com/leondgarse/keras_cv_attention_models).
Keras implementation of ResNetD.
Paper [PDF 1812.01187 Bag of Tricks for Image Classification with Convolutional Neural Networks](https://arxiv.org/pdf/1812.01187.pdf).
"""

ResNetD.__doc__ = __resnetd_head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  deep_stem: Boolean value if use deep stem. Default `True`.
  stem_width: output dimension for stem block. Default `32`.
  strides: a `number` or `list`, indicates strides used in the last stack or list value for all stacks.
      If a number, it will be `[1, 2, 2, strides]`. Default `2`.
  model_name: string, model name.
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"]) + """
Model architectures:
  | Model      | Params | FLOPs  | Input | Top1 Acc |
  | ---------- | ------ | ------ | ----- | -------- |
  | ResNet50D  | 25.58M | 4.33G  | 224   | 80.530   |
  | ResNet101D | 44.57M | 8.04G  | 224   | 83.022   |
  | ResNet152D | 60.21M | 11.75G | 224   | 83.680   |
  | ResNet200D | 64.69M | 15.25G | 224   | 83.962   |
"""

__resnetd_default_doc__ = __resnetd_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"])

ResNet50D.__doc__ = __resnetd_default_doc__
ResNet101D.__doc__ = __resnetd_default_doc__
ResNet152D.__doc__ = __resnetd_default_doc__
ResNet200D.__doc__ = __resnetd_default_doc__

__resnetq_head_doc__ = """
Github source [leondgarse/keras_cv_attention_models](https://github.com/leondgarse/keras_cv_attention_models).
Defined and model weights loaded from [Github timm/models/resnet.py](https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/resnet.py).
"""

__resnetq_tail_doc__ = """  strides: a `number` or `list`, indicates strides used in the last stack or list value for all stacks.
      If a number, it will be `[1, 2, 2, strides]`. Default `2`.
  stem_downsample: Boolean value if add `MaxPooling2D` layer after stem block.
  input_shape: it should have exactly 3 inputs channels, default `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `softmax`.
  pretrained: value in [None, "imagenet"].
      Will try to download and load pre-trained model weights if not None.
      Save path is `~/.keras/models/`.

Returns:
    A `keras.Model` instance.
"""

ResNetQ.__doc__ = __resnetq_head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: Output channel for each stack.
  stem_width: output dimension for stem block.
  stem_act: Boolean value if add use `batchnorm + activation` in stem branch.
  expansion: filter expansion in each block. The larger the wider.
  groups_div: number value or list of number, controls `groups` in each stack.
      - `0` means using `edge block`, which removes the first `1x1` conv.
      - `(0, 1)` values means `groups=filter // groups_div` for `Conv2D`.
      - `1` means `groups == filter`, will apply `DepthwiseConv2D`.
  extra_conv: Boolean value or list of Boolean, if add another `3x3` conv in each block.
  num_features: none `0` value to add another `conv2d + bn + activation` layers before `GlobalAveragePooling2D`.
  model_name: string, model name.
""" + __resnetq_tail_doc__ + """
Model architectures:
  | Model     | Params | FLOPs | Input | Top1 Acc |
  | --------- | ------ | ----- | ----- | -------- |
  | ResNet51Q | 35.7M  | 4.87G | 224   | 82.36    |
  | ResNet61Q | 36.8M  | 5.96G | 224   |          |
"""

ResNet51Q.__doc__ = __resnetq_head_doc__ + """
Args:
""" + __resnetq_tail_doc__
ResNet61Q.__doc__ = __resnetq_head_doc__ + """
Args:
""" + __resnetq_tail_doc__

__regnety_head_doc__ = """
Keras implementation of [Github facebookresearch/regnet](https://github.com/facebookresearch/pycls/blob/main/pycls/models/regnet.py).
Paper [PDF 2003.13678 Designing Network Design Spaces](https://arxiv.org/pdf/2003.13678.pdf).
Model weights loaded from [timm/regnet](https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/regnet.py).
Paper [PDF 2110.00476 ResNet strikes back: An improved training procedure in timm](https://arxiv.org/pdf/2110.00476.pdf).
"""

RegNetY.__doc__ = __regnety_head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: Output channel for each stack.
  model_name: string, model name.
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"]) + """
Model architectures:
  | Model      | Params  | FLOPs  | Input | Top1 Acc |
  | ---------- | ------- | ------ | ----- | -------- |
  | RegNetY040 | 20.65M  | 3.98G  | 224   | 82.3     |
  | RegNetY064 | 30.58M  | 6.36G  | 224   | 83.0     |
  | RegNetY080 | 39.18M  | 7.97G  | 224   | 83.17    |
  | RegNetY160 | 83.59M  | 15.92G | 224   | 82.0     |
  | RegNetY320 | 145.05M | 32.29G | 224   | 82.5     |
"""

__regnety_default_doc__ = __regnety_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"])

RegNetY040.__doc__ = __regnety_default_doc__
RegNetY064.__doc__ = __regnety_default_doc__
RegNetY080.__doc__ = __regnety_default_doc__
RegNetY160.__doc__ = __regnety_default_doc__
RegNetY320.__doc__ = __regnety_default_doc__

__regnetz_head_doc__ = """
Github source [leondgarse/keras_cv_attention_models](https://github.com/leondgarse/keras_cv_attention_models).
Defined and model weights loaded from [timm](https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/byobnet.py).
Related paper [PDF 2004.02967 Evolving Normalization-Activation Layers](https://arxiv.org/pdf/2004.02967.pdf).
"""

__regnetz_default_doc__ = __regnetz_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"]) + """
Model architectures:
  | Model          | Params | FLOPs | Input | Top1 Acc |
  | -------------- | ------ | ----- | ----- | -------- |
  | RegNetZB16     | 9.72M  | 1.44G | 224   | 79.868   |
  | RegNetZC16     | 13.46M | 2.50G | 256   | 82.164   |
  | RegNetZC16_EVO | 13.49M | 2.55G | 256   | 81.9     |
  | RegNetZD32     | 27.58M | 5.96G | 256   | 83.422   |
  | RegNetZD8      | 23.37M | 3.95G | 256   | 83.5     |
  | RegNetZD8_EVO  | 23.46M | 4.61G | 256   | 83.42    |
  | RegNetZE8      | 57.70M | 9.88G | 256   | 84.5     |
"""

RegNetZB16.__doc__ = __regnetz_default_doc__
RegNetZC16.__doc__ = __regnetz_default_doc__
RegNetZC16_EVO.__doc__ = __regnetz_default_doc__
RegNetZD32.__doc__ = __regnetz_default_doc__
RegNetZD8.__doc__ = __regnetz_default_doc__
RegNetZD8_EVO.__doc__ = __regnetz_default_doc__
RegNetZE8.__doc__ = __regnetz_default_doc__



import math
import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, initializers
from keras_cv_attention_models.attention_layers import (
    BiasLayer,
    # ChannelAffine,
    drop_block,
    layer_norm,
    mlp_block,
    output_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "swin_transformer_v2_base_window12": {"imagenet21k": {192: "b68add0f06aeede5d4dd10cd26855f36"}},
    "swin_transformer_v2_base_window16": {"imagenet22k": {256: "1caba96ed702d467e650465503033c85"}},
    "swin_transformer_v2_base_window16": {"imagenet": {256: "a7bfb9ae0733807baa702d05eb8feab8"}},
    "swin_transformer_v2_base_window24": {"imagenet22k": {384: "acd467e1d8555e15542c8254bdae1b72"}},
    "swin_transformer_v2_base_window8": {"imagenet": {256: "15454d9f6ba2ccca940f9c45b6935af6"}},
    "swin_transformer_v2_large_window12": {"imagenet21k": {192: "ace20b0d634eb92989ece52e300440d5"}},
    "swin_transformer_v2_large_window16": {"imagenet22k": {256: "151bb82a138f956613ce4b9885bfdd18"}},
    "swin_transformer_v2_large_window24": {"imagenet22k": {384: "04fa3b195e5201c2d1068d1e19c8a0c5"}},
    "swin_transformer_v2_small_window16": {"imagenet": {256: "3b2ca43d1927cca1b414b60e1044a84d"}},
    "swin_transformer_v2_small_window8": {"imagenet": {256: "0a8468bd9acdf2056fc401e9f5067f97"}},
    "swin_transformer_v2_tiny_window16": {"imagenet": {256: "37ce8c5f514c2249ef10d9c3acc37d29"}},
    "swin_transformer_v2_tiny_window8": {"imagenet": {256: "9317f155e37e4081a09d290ca99bf7cd"}},
}


@backend.register_keras_serializable(package="kecam")
class ExpLogitScale(layers.Layer):
    def __init__(self, axis=-1, init_value=10.0, max_value=100.0, **kwargs):
        super().__init__(**kwargs)
        self.axis, self.init_value, self.max_value = axis, init_value, max_value

    def build(self, input_shape):
        if self.axis == -1 or self.axis == len(input_shape) - 1:
            weight_shape = (input_shape[-1],)
        else:
            weight_shape = [1] * len(input_shape)
            axis = self.axis if isinstance(self.axis, (list, tuple)) else [self.axis]
            for ii in axis:
                weight_shape[ii] = input_shape[ii]

        initializer = initializers.constant(math.log(self.init_value))
        self.scale = self.add_weight(name="weight", shape=weight_shape, initializer=initializer, trainable=True, dtype="float32")
        # self.__max_value__ = functional.convert_to_tensor(float(math.log(self.max_value)))
        self.__max_value__ = float(math.log(self.max_value))
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        return inputs * functional.exp(functional.minimum(self.scale, self.__max_value__))

    def get_config(self):
        config = super().get_config()
        config.update({"axis": self.axis, "max_value": self.max_value})
        return config


@backend.register_keras_serializable(package="kecam")
class PairWiseRelativePositionalEmbedding(layers.Layer):
    def __init__(self, pos_scale=-1, **kwargs):
        # No weight, just need to wrapper a layer, or will not in model structure
        super().__init__(**kwargs)
        self.pos_scale = pos_scale
        self.use_layer_as_module = True

    def build(self, input_shape):
        # input_shape: [batch * window_patch, window_height, window_width, channel]
        height, width = input_shape[1], input_shape[2]  # [12, 15]
        hh, ww = np.meshgrid(range(-height + 1, height), range(-width + 1, width), indexing="ij")
        coords = np.stack([hh, ww], axis=-1).astype("float32")
        if self.pos_scale == -1:
            pos_scale = [height, width]
        else:
            # If pretrined weights are from different input_shape or window_size, pos_scale is previous actually using window_size
            pos_scale = self.pos_scale if isinstance(self.pos_scale, (list, tuple)) else [self.pos_scale, self.pos_scale]
        coords = coords * 8 / [float(pos_scale[0] - 1), float(pos_scale[1] - 1)]  # [23, 29, 2], normalize to -8, 8
        # torch.sign(relative_coords_table) * torch.log2(torch.abs(relative_coords_table) + 1.0) / math.log2(8)
        relative_log_coords = np.sign(coords) * np.log(1.0 + np.abs(coords)) / (np.log(2.0) * 3.0)
        relative_log_coords = np.reshape(relative_log_coords, [-1, 2])  # [23 * 29, 2]
        self.relative_log_coords = functional.convert_to_tensor(relative_log_coords, dtype="float32")
        self.height, self.width = height, width  # For reload with shape mismatched
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        return self.relative_log_coords

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"pos_scale": self.pos_scale})
        return base_config

    def compute_output_shape(self, input_shape):
        return [(2 * self.height - 1) * (2 * self.width - 1), 2]


@backend.register_keras_serializable(package="kecam")
class PairWiseRelativePositionalEmbeddingGather(layers.Layer):
    def __init__(self, window_height=-1, **kwargs):
        # No weight, just need to wrapper a layer, or will not in model structure
        super().__init__(**kwargs)
        self.window_height = window_height
        self.use_layer_as_module = True

    def build(self, input_shape):
        # input_shape: [(2 * window_height - 1) * (2 * window_width - 1), num_heads]
        height = self.window_height if self.window_height > 0 else int(np.sqrt(input_shape[0]) + 1) // 2
        width = ((input_shape[0] // (2 * height - 1)) + 1) // 2

        hh, ww = np.meshgrid(range(height), range(width))
        coords = np.stack([hh, ww], axis=-1).astype("float32")  # [15, 12, 2]
        coords_flatten = np.reshape(coords, [-1, 2])  # [180, 2]
        relative_coords = coords_flatten[:, None, :] - coords_flatten[None, :, :]  # [180, 180, 2]
        # relative_coords = tf.reshape(relative_coords, [-1, 2])  # [196 * 196, 2]

        relative_coords_hh = relative_coords[:, :, 0] + height - 1
        relative_coords_ww = (relative_coords[:, :, 1] + width - 1) * (2 * height - 1)
        relative_coords_hhww = np.stack([relative_coords_hh, relative_coords_ww], axis=-1)
        relative_position_index = np.sum(relative_coords_hhww, axis=-1)  # [180, 180]
        self.relative_position_index = functional.convert_to_tensor(relative_position_index, dtype="int64")
        self.height, self.width, self.num_heads = height, width, input_shape[-1]
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        # return self.relative_position_index
        return functional.gather(inputs, self.relative_position_index)  # [hh * ww, hh * ww, num_heads]

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"window_height": self.window_height})
        return base_config

    def compute_output_shape(self, input_shape):
        return [self.height * self.width, self.height * self.width, self.num_heads]


@backend.register_keras_serializable(package="kecam")
class WindowAttentionMask(layers.Layer):
    def __init__(self, height, width, window_height, window_width, shift_height=0, shift_width=0, **kwargs):
        # No weight, just need to wrapper a layer, or will meet some error in model saving or loading...
        self.height, self.width, self.window_height, self.window_width = height, width, window_height, window_width
        self.shift_height, self.shift_width = shift_height, shift_width
        self.blocks = (self.height // self.window_height) * (self.width // self.window_width)
        super().__init__(**kwargs)

    def build(self, input_shape):
        hh_split = [0, self.height - self.window_height, self.height - self.shift_height, self.height]
        ww_split = [0, self.width - self.window_width, self.width - self.shift_width, self.width]
        mask_value, total_ww, mask = 0, len(ww_split) - 1, []
        for hh_id in range(len(hh_split) - 1):
            hh = hh_split[hh_id + 1] - hh_split[hh_id]
            rr = [np.zeros([hh, ww_split[id + 1] - ww_split[id]], dtype="float32") + (id + mask_value) for id in range(total_ww)]
            mask.append(np.concatenate(rr, axis=-1))
            mask_value += total_ww
        mask = np.concatenate(mask, axis=0)
        # return mask

        mask = np.reshape(mask, [self.height // self.window_height, self.window_height, self.width // self.window_width, self.window_width])
        mask = np.transpose(mask, [0, 2, 1, 3])
        mask = np.reshape(mask, [-1, self.window_height * self.window_width])
        attn_mask = np.expand_dims(mask, 1) - np.expand_dims(mask, 2)
        # attn_mask = tf.cast(np.where(attn_mask != 0, -100, 0), self._compute_dtype)
        attn_mask = np.where(attn_mask != 0, -100, 0)
        attn_mask = np.expand_dims(np.expand_dims(attn_mask, 1), 0)  # expand dims on batch and num_heads
        # attn_mask = functional.convert_to_tensor(attn_mask, dtype="float32")

        if hasattr(self, "register_buffer"):  # PyTorch
            self.register_buffer("attn_mask", functional.convert_to_tensor(attn_mask, dtype="float32"), persistent=False)
        else:
            self.attn_mask = functional.convert_to_tensor(attn_mask, dtype="float32")

        self.num_heads, self.query_blocks = input_shape[1], input_shape[2]
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        # inputs: [batch_size * blocks, num_heads, query_blocks, query_blocks]
        # where query_blocks = `window_height * window_width`, blocks = `(height // window_height) * (width // window_width)`
        nn = functional.reshape(inputs, [-1, self.blocks, self.num_heads, self.query_blocks, self.query_blocks])
        nn = nn + self.attn_mask
        return functional.reshape(nn, [-1, self.num_heads, self.query_blocks, self.query_blocks])

    def compute_output_shape(self, input_shape):
        return [None, self.num_heads, self.query_blocks, self.query_blocks]

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "height": self.height,
                "width": self.width,
                "window_height": self.window_height,
                "window_width": self.window_width,
                "shift_height": self.shift_height,
                "shift_width": self.shift_width,
            }
        )
        return config


def window_mhsa_with_pair_wise_positional_embedding(
    inputs, num_heads=4, key_dim=0, meta_hidden_dim=512, mask=None, pos_scale=-1, out_bias=True, qv_bias=True, attn_dropout=0, out_dropout=0, name=None
):
    input_channel = inputs.shape[-1]
    key_dim = key_dim if key_dim > 0 else input_channel // num_heads
    qk_out = key_dim * num_heads

    qkv = layers.Dense(qk_out * 3, use_bias=False, name=name and name + "qkv")(inputs)
    qkv = functional.reshape(qkv, [-1, qkv.shape[1] * qkv.shape[2], qkv.shape[-1]])
    query, key, value = functional.split(qkv, 3, axis=-1)
    if qv_bias:
        query = BiasLayer(name=name and name + "query_bias")(query)
        value = BiasLayer(name=name and name + "value_bias")(value)
    query = functional.transpose(functional.reshape(query, [-1, query.shape[1], num_heads, key_dim]), [0, 2, 1, 3])  # [batch, num_heads, hh * ww, key_dim]
    key = functional.transpose(functional.reshape(key, [-1, key.shape[1], num_heads, key_dim]), [0, 2, 3, 1])  # [batch, num_heads, key_dim, hh * ww]
    value = functional.transpose(functional.reshape(value, [-1, value.shape[1], num_heads, key_dim]), [0, 2, 1, 3])  # [batch, num_heads, hh * ww, vv_dim]

    # cosine attention
    norm_query, norm_key = functional.l2_normalize(query, axis=-1, epsilon=1e-6), functional.l2_normalize(key, axis=-2, epsilon=1e-6)
    attn = functional.matmul(norm_query, norm_key)  # [batch, num_heads, hh * ww, hh * ww]
    attn = ExpLogitScale(axis=1, name=name and name + "scale")(attn)  # axis=1 is head dimension

    # PairWiseRelativePositionalEmbedding -> mlp -> add with attn
    pos_coords = PairWiseRelativePositionalEmbedding(pos_scale=pos_scale, name=name and name + "pos_emb")(inputs)
    relative_position_bias = layers.Dense(meta_hidden_dim, use_bias=True, name=name and name + "meta_dense_1")(pos_coords)
    relative_position_bias = layers.Activation("relu")(relative_position_bias)
    relative_position_bias = layers.Dense(num_heads, use_bias=False, name=name and name + "meta_dense_2")(relative_position_bias)

    # relative_position_bias = functional.gather(relative_position_bias, pos_index)  # [hh * ww, hh * ww, num_heads]
    relative_position_bias = PairWiseRelativePositionalEmbeddingGather(window_height=inputs.shape[1], name=name and name + "pos_gather")(relative_position_bias)
    relative_position_bias = functional.sigmoid(relative_position_bias) * 16.0
    relative_position_bias = functional.expand_dims(functional.transpose(relative_position_bias, [2, 0, 1]), 0)  # [1, num_heads, hh * ww, hh * ww]
    attn = attn + relative_position_bias

    if mask is not None:
        attn = mask(attn)
    attention_scores = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attn)

    if attn_dropout > 0:
        attention_scores = layers.Dropout(attn_dropout, name=name and name + "attn_drop")(attention_scores)
    attention_output = functional.matmul(attention_scores, value)
    attention_output = functional.transpose(attention_output, perm=[0, 2, 1, 3])
    attention_output = functional.reshape(attention_output, [-1, inputs.shape[1], inputs.shape[2], num_heads * key_dim])
    # print(f">>>> {attention_output.shape = }, {attention_scores.shape = }")

    # [batch, hh, ww, num_heads * vv_dim] * [num_heads * vv_dim, out] --> [batch, hh, ww, out]
    attention_output = layers.Dense(qk_out, use_bias=out_bias, name=name and name + "output")(attention_output)
    attention_output = layers.Dropout(out_dropout, name=name and name + "out_drop")(attention_output) if out_dropout > 0 else attention_output
    return attention_output


def shifted_window_attention(inputs, window_size, num_heads=4, shift_size=0, pos_scale=-1, name=None):
    input_channel = inputs.shape[-1]
    window_size = window_size if isinstance(window_size, (list, tuple)) else [window_size, window_size]
    window_height = window_size[0] if window_size[0] < inputs.shape[1] else inputs.shape[1]
    window_width = window_size[1] if window_size[1] < inputs.shape[2] else inputs.shape[2]
    shift_size = 0 if (window_height == inputs.shape[1] and window_width == inputs.shape[2]) else shift_size
    should_shift = shift_size > 0

    # window_partition, partition windows, ceil mode padding if not divisible by window_size
    # patch_height, patch_width = inputs.shape[1] // window_height, inputs.shape[2] // window_width
    patch_height, patch_width = int(math.ceil(inputs.shape[1] / window_height)), int(math.ceil(inputs.shape[2] / window_width))
    should_pad_hh, should_pad_ww = patch_height * window_height - inputs.shape[1], patch_width * window_width - inputs.shape[2]
    # print(f">>>> shifted_window_attention {inputs.shape = }, {should_pad_hh = }, {should_pad_ww = }")
    if should_pad_hh or should_pad_ww:
        inputs = functional.pad(inputs, [[0, 0], [0, should_pad_hh], [0, should_pad_ww], [0, 0]])

    if should_shift:
        shift_height, shift_width = int(window_height * shift_size), int(window_width * shift_size)
        # functional.roll is not supported by tflite
        # inputs = functional.roll(inputs, shift=(shift_height * -1, shift_width * -1), axis=[1, 2])
        inputs = functional.concat([inputs[:, shift_height:], inputs[:, :shift_height]], axis=1)
        inputs = functional.concat([inputs[:, :, shift_width:], inputs[:, :, :shift_width]], axis=2)

    # print(f">>>> shifted_window_attention {inputs.shape = }, {patch_height = }, {patch_width = }, {window_height = }, {window_width = }")
    # [batch * patch_height, window_height, patch_width, window_width * channels], limit transpose perm <= 4
    nn = functional.reshape(inputs, [-1, window_height, patch_width, window_width * input_channel])
    nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch * patch_height, patch_width, window_height, window_width * channels]
    nn = functional.reshape(nn, [-1, window_height, window_width, input_channel])  # [batch * patch_height * patch_width, window_height, window_width, channels]

    mask = WindowAttentionMask(inputs.shape[1], inputs.shape[2], window_height, window_width, shift_height, shift_width) if should_shift else None
    nn = window_mhsa_with_pair_wise_positional_embedding(nn, num_heads=num_heads, mask=mask, pos_scale=pos_scale, name=name)

    # window_reverse, merge windows
    # [batch * patch_height, patch_width, window_height, window_width * input_channel], limit transpose perm <= 4
    nn = functional.reshape(nn, [-1, patch_width, window_height, window_width * input_channel])
    nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch * patch_height, window_height, patch_width, window_width * input_channel]
    nn = functional.reshape(nn, [-1, patch_height * window_height, patch_width * window_width, input_channel])

    if should_shift:
        # nn = functional.roll(nn, shift=(shift_height, shift_width), axis=[1, 2])
        nn = functional.concat([nn[:, -shift_height:], nn[:, :-shift_height]], axis=1)
        nn = functional.concat([nn[:, :, -shift_width:], nn[:, :, :-shift_width]], axis=2)

    # print(f">>>> shifted_window_attention before: {nn.shape = }, {should_pad_hh = }, {should_pad_ww = }")
    if should_pad_hh or should_pad_ww:
        nn = nn[:, : nn.shape[1] - should_pad_hh, : nn.shape[2] - should_pad_ww, :]  # In case should_pad_hh or should_pad_ww is 0
    # print(f">>>> shifted_window_attention after: {nn.shape = }")

    return nn


def swin_transformer_block(
    inputs, window_size, num_heads=4, shift_size=0, pos_scale=-1, mlp_ratio=4, mlp_drop_rate=0, attn_drop_rate=0, drop_rate=0, name=None
):
    input_channel = inputs.shape[-1]
    attn = shifted_window_attention(inputs, window_size, num_heads, shift_size, pos_scale=pos_scale, name=name + "attn_")
    attn = layer_norm(attn, zero_gamma=True, axis=-1, name=name + "attn_")
    # attn = ChannelAffine(use_bias=False, weight_init_value=layer_scale, name=name + "1_gamma")(attn) if layer_scale >= 0 else attn
    attn = drop_block(attn, drop_rate=drop_rate, name=name + "attn_")
    attn_out = layers.Add(name=name + "attn_out")([inputs, attn])

    mlp = mlp_block(attn_out, int(input_channel * mlp_ratio), drop_rate=mlp_drop_rate, use_conv=False, activation="gelu", name=name + "mlp_")
    mlp = layer_norm(mlp, zero_gamma=True, axis=-1, name=name + "mlp_")
    # mlp = ChannelAffine(use_bias=False, weight_init_value=layer_scale, name=name + "2_gamma")(mlp) if layer_scale >= 0 else mlp
    mlp = drop_block(mlp, drop_rate=drop_rate, name=name + "mlp_")
    return layers.Add(name=name + "output")([attn_out, mlp])


def patch_merging(inputs, name=""):
    input_channel = inputs.shape[-1]
    should_pad_hh, should_pad_ww = inputs.shape[1] % 2, inputs.shape[2] % 2
    # print(f">>>> patch_merging {inputs.shape = }, {should_pad_hh = }, {should_pad_ww = }")
    if should_pad_hh or should_pad_ww:
        inputs = functional.pad(inputs, [[0, 0], [0, should_pad_hh], [0, should_pad_ww], [0, 0]])

    # limit transpose perm <= 4
    nn = functional.reshape(inputs, [-1, 2, inputs.shape[2], input_channel])  # [batch * inputs.shape[1] // 2, height 2, inputs.shape[2], input_channel]
    nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch * inputs.shape[1] // 2, inputs.shape[2], height 2, input_channel]
    nn = functional.reshape(nn, [-1, inputs.shape[1] // 2, inputs.shape[2] // 2, 2 * 2 * input_channel])
    nn = layers.Dense(2 * input_channel, use_bias=False, name=name + "dense")(nn)
    nn = layer_norm(nn, axis=-1, name=name)
    return nn


def SwinTransformerV2(
    num_blocks=[2, 2, 6, 2],
    num_heads=[3, 6, 12, 24],
    embed_dim=96,
    window_size=7,
    pos_scale=-1,  # If pretrained weights are from different input_shape or window_size, pos_scale is previous actually using window_size
    stem_patch_size=4,
    use_stack_norm=False,  # True for extra layer_norm on each stack end
    extra_norm_period=0,  # > 0 for extra layer_norm frequency in each stack. May combine with use_stack_norm=True
    input_shape=(224, 224, 3),
    num_classes=1000,
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="swin_transformer_v2",
    kwargs=None,
):
    """Patch stem"""
    # Regard input_shape as force using original shape if first element is None or -1,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    nn = layers.Conv2D(embed_dim, kernel_size=stem_patch_size, strides=stem_patch_size, use_bias=True, name="stem_conv")(inputs)
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)  # channels_first -> channels_last
    nn = layer_norm(nn, axis=-1, name="stem_")
    window_size = window_size if isinstance(window_size, (list, tuple)) else [window_size, window_size]

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, num_head) in enumerate(zip(num_blocks, num_heads)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            # height, width downsample * 0.5, channel upsample * 2
            nn = patch_merging(nn, name=stack_name + "downsample_")
        cur_pos_scale = pos_scale[stack_id] if isinstance(pos_scale, (list, tuple)) else pos_scale
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            shift_size = 0 if block_id % 2 == 0 else 0.5
            nn = swin_transformer_block(nn, window_size, num_head, shift_size, cur_pos_scale, drop_rate=block_drop_rate, name=block_name)
            global_block_id += 1
            if extra_norm_period > 0 and (block_id + 1) % extra_norm_period == 0 and not (use_stack_norm and block_id == num_block - 1):
                nn = layer_norm(nn, axis=-1, name=block_name + "output_")
        if use_stack_norm and stack_id != len(num_blocks) - 1:  # Exclude last stack
            nn = layer_norm(nn, axis=-1, name=stack_name + "output_")
    nn = layer_norm(nn, axis=-1, name="pre_output_")
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first

    nn = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "swin_transformer_v2", pretrained)
    return model


def SwinTransformerV2Tiny_window8(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    window_size = kwargs.pop("window_size", 8)
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_tiny_window8", **kwargs)


def SwinTransformerV2Tiny_window16(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    window_size = kwargs.pop("window_size", 16)
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_tiny_window16", **kwargs)


def SwinTransformerV2Small_window8(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 18, 2]
    window_size = kwargs.pop("window_size", 8)
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_small_window8", **kwargs)


def SwinTransformerV2Small_window16(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 18, 2]
    window_size = kwargs.pop("window_size", 16)
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_small_window16", **kwargs)


def SwinTransformerV2Base_window8(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 18, 2]
    num_heads = [4, 8, 16, 32]
    embed_dim = 128
    window_size = kwargs.pop("window_size", 8)
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_base_window8", **kwargs)


def SwinTransformerV2Base_window12(input_shape=(192, 192, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet21k", **kwargs):
    num_blocks = [2, 2, 18, 2]
    num_heads = [4, 8, 16, 32]
    embed_dim = 128
    window_size = kwargs.pop("window_size", 12)
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_base_window12", **kwargs)


def SwinTransformerV2Base_window16(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 18, 2]
    num_heads = [4, 8, 16, 32]
    embed_dim = 128
    window_size = kwargs.pop("window_size", 16)
    pos_scale = kwargs.pop("pos_scale", [12, 12, 12, 6] if pretrained == "imagenet22k" else -1)  # 22k model is fine-tuned from window12
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_base_window16", **kwargs)


def SwinTransformerV2Base_window24(input_shape=(384, 384, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet22k", **kwargs):
    num_blocks = [2, 2, 18, 2]
    num_heads = [4, 8, 16, 32]
    embed_dim = 128
    window_size = kwargs.pop("window_size", 24)
    pos_scale = kwargs.pop("pos_scale", [12, 12, 12, 6] if pretrained == "imagenet22k" else -1)  # 22k model is fine-tuned from window12
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_base_window24", **kwargs)


def SwinTransformerV2Large_window12(input_shape=(192, 192, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet21k", **kwargs):
    num_blocks = [2, 2, 18, 2]
    num_heads = [6, 12, 24, 48]
    embed_dim = 192
    window_size = kwargs.pop("window_size", 12)
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_large_window12", **kwargs)


def SwinTransformerV2Large_window16(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet22k", **kwargs):
    num_blocks = [2, 2, 18, 2]
    num_heads = [6, 12, 24, 48]
    embed_dim = 192
    window_size = kwargs.pop("window_size", 16)
    pos_scale = kwargs.pop("pos_scale", [12, 12, 12, 6] if pretrained == "imagenet22k" else -1)  # 22k model is fine-tuned from window12
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_large_window16", **kwargs)


def SwinTransformerV2Large_window24(input_shape=(384, 384, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet22k", **kwargs):
    num_blocks = [2, 2, 18, 2]
    num_heads = [6, 12, 24, 48]
    embed_dim = 192
    window_size = kwargs.pop("window_size", 24)
    pos_scale = kwargs.pop("pos_scale", [12, 12, 12, 6] if pretrained == "imagenet22k" else -1)  # 22k model is fine-tuned from window12
    return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_large_window24", **kwargs)


# def SwinTransformerV2Giant(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained=None, **kwargs):
#     num_blocks = [2, 2, 42, 2]
#     num_heads = [16, 32, 64, 128]
#     embed_dim = 512
#     extra_norm_period = 6
#     return SwinTransformerV2(**locals(), model_name="swin_transformer_v2_giant", **kwargs)
